"""
Grains from cloud metadata servers at 169.254.169.254 in Azure Virtual Machine

.. versionadded:: 3006.0

:depends: requests

To enable these grains that pull from the http://169.254.169.254/metadata/instance?api-version=2020-09-01
metadata server set `metadata_server_grains: True` in the minion config.

.. code-block:: yaml

    metadata_server_grains: True

"""

import logging

import salt.utils.http as http
import salt.utils.json

HOST = "http://169.254.169.254"
URL = f"{HOST}/metadata/instance?api-version=2020-09-01"
log = logging.getLogger(__name__)


def __virtual__():
    # Check if metadata_server_grains minion option is enabled
    if __opts__.get("metadata_server_grains", False) is False:
        return False
    azuretest = http.query(
        URL, status=True, headers=True, header_list=["Metadata: true"]
    )
    if azuretest.get("status", 404) != 200:
        return False
    return True


def metadata():
    """Takes no arguments, returns a dictionary of metadata values from Azure."""
    log.debug("All checks true - loading azure metadata")
    result = http.query(URL, headers=True, header_list=["Metadata: true"])
    metadata = salt.utils.json.loads(result.get("body", {}))

    return metadata



"""
Management of Influxdb retention policies
=========================================

.. versionadded:: 2017.7.0

(compatible with InfluxDB version 0.9+)
"""


def __virtual__():
    """
    Only load if the influxdb module is available
    """
    if "influxdb.db_exists" in __salt__:
        return "influxdb_retention_policy"
    return (False, "influxdb module could not be loaded")


def convert_duration(duration):
    """
    Convert the a duration string into XXhYYmZZs format

    duration
        Duration to convert

    Returns: duration_string
        String representation of duration in XXhYYmZZs format
    """

    # durations must be specified in days, weeks or hours

    if duration.endswith("h"):
        hours = int(duration.split("h"))

    elif duration.endswith("d"):
        days = duration.split("d")
        hours = int(days[0]) * 24

    elif duration.endswith("w"):
        weeks = duration.split("w")
        hours = int(weeks[0]) * 24 * 7

    duration_string = str(hours) + "h0m0s"
    return duration_string


def present(name, database, duration="7d", replication=1, default=False, **client_args):
    """
    Ensure that given retention policy is present.

    name
        Name of the retention policy to create.

    database
        Database to create retention policy on.
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "retention policy {} is already present".format(name),
    }

    if not __salt__["influxdb.retention_policy_exists"](
        name=name, database=database, **client_args
    ):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = " {} is absent and will be created".format(name)
            return ret
        if __salt__["influxdb.create_retention_policy"](
            database, name, duration, replication, default, **client_args
        ):
            ret["comment"] = "retention policy {} has been created".format(name)
            ret["changes"][name] = "Present"
            return ret
        else:
            ret["comment"] = "Failed to create retention policy {}".format(name)
            ret["result"] = False
            return ret

    else:
        current_policy = __salt__["influxdb.get_retention_policy"](
            database=database, name=name, **client_args
        )
        update_policy = False
        if current_policy["duration"] != convert_duration(duration):
            update_policy = True
            ret["changes"]["duration"] = "Retention changed from {} to {}.".format(
                current_policy["duration"], duration
            )

        if current_policy["replicaN"] != replication:
            update_policy = True
            ret["changes"]["replication"] = "Replication changed from {} to {}.".format(
                current_policy["replicaN"], replication
            )

        if current_policy["default"] != default:
            update_policy = True
            ret["changes"]["default"] = "Default changed from {} to {}.".format(
                current_policy["default"], default
            )

        if update_policy:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = " {} is present and set to be changed".format(name)
                return ret
            else:
                if __salt__["influxdb.alter_retention_policy"](
                    database, name, duration, replication, default, **client_args
                ):
                    ret["comment"] = "retention policy {} has been changed".format(name)
                    return ret
                else:
                    ret["comment"] = "Failed to update retention policy {}".format(name)
                    ret["result"] = False
                    return ret

    return ret


def absent(name, database, **client_args):
    """
    Ensure that given retention policy is absent.

    name
        Name of the retention policy to remove.

    database
        Name of the database that the retention policy was defined on.
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "retention policy {} is not present".format(name),
    }

    if __salt__["influxdb.retention_policy_exists"](database, name, **client_args):
        if __opts__["test"]:
            ret["result"] = None
            ret[
                "comment"
            ] = "retention policy {} is present and needs to be removed".format(name)
            return ret
        if __salt__["influxdb.drop_retention_policy"](database, name, **client_args):
            ret["comment"] = "retention policy {} has been removed".format(name)
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["comment"] = "Failed to remove retention policy {}".format(name)
            ret["result"] = False
            return ret

    return ret



"""
Powerpath configuration support
===============================

Allows configuration of EMC Powerpath.  Currently
only addition/deletion of licenses is supported.

.. code-block:: yaml

    key:
      powerpath.license_present: []
"""


def license_present(name):
    """
    Ensures that the specified PowerPath license key is present
    on the host.

    name
        The license key to ensure is present
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if not __salt__["powerpath.has_powerpath"]():
        ret["result"] = False
        ret["comment"] = "PowerPath is not installed."
        return ret

    licenses = [l["key"] for l in __salt__["powerpath.list_licenses"]()]

    if name in licenses:
        ret["result"] = True
        ret["comment"] = "License key {} already present".format(name)
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "License key {} is set to be added".format(name)
        return ret

    data = __salt__["powerpath.add_license"](name)
    if data["result"]:
        ret["changes"] = {name: "added"}
        ret["result"] = True
        ret["comment"] = data["output"]
        return ret
    else:
        ret["result"] = False
        ret["comment"] = data["output"]
        return ret


def license_absent(name):
    """
    Ensures that the specified PowerPath license key is absent
    on the host.

    name
        The license key to ensure is absent
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if not __salt__["powerpath.has_powerpath"]():
        ret["result"] = False
        ret["comment"] = "PowerPath is not installed."
        return ret

    licenses = [l["key"] for l in __salt__["powerpath.list_licenses"]()]

    if name not in licenses:
        ret["result"] = True
        ret["comment"] = "License key {} not present".format(name)
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "License key {} is set to be removed".format(name)
        return ret

    data = __salt__["powerpath.remove_license"](name)
    if data["result"]:
        ret["changes"] = {name: "removed"}
        ret["result"] = True
        ret["comment"] = data["output"]
        return ret
    else:
        ret["result"] = False
        ret["comment"] = data["output"]
        return ret



import logging

import pytest

import salt.cache
import salt.loader
from tests.pytests.functional.cache.helpers import run_common_cache_tests
from tests.support.pytest.mysql import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


@pytest.fixture(scope="module")
def mysql_combo(create_mysql_combo):  # pylint: disable=function-redefined
    create_mysql_combo.mysql_database = "salt_cache"
    return create_mysql_combo


@pytest.fixture
def cache(minion_opts, mysql_container):
    opts = minion_opts.copy()
    opts["cache"] = "mysql"
    opts["mysql.host"] = "127.0.0.1"
    opts["mysql.port"] = mysql_container.mysql_port
    opts["mysql.user"] = mysql_container.mysql_user
    opts["mysql.password"] = mysql_container.mysql_passwd
    opts["mysql.database"] = mysql_container.mysql_database
    opts["mysql.table_name"] = "cache"
    cache = salt.cache.factory(opts)
    return cache


def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)



from types import SimpleNamespace

import pytest

import salt.netapi.rest_cherrypy.app as cherrypy_app
from tests.support.mock import MagicMock, patch


class MockCherryPy:
    session = MagicMock(cache={}, id="6d1b722e")
    config = {
        "saltopts": {},
        "apiopts": {
            "external_auth": {"rest": {"^url": "https://test_url/rest"}},
            "cachedir": "/tmp",
        },
    }
    request = SimpleNamespace(
        lowstate=[{"username": "fred", "password": "secret"}],
        remote=SimpleNamespace(ip="1.2.3.4"),
    )
    serving = SimpleNamespace(request=request)
    response = SimpleNamespace(headers={})


class MockNetapiClient:
    def __init__(self, *args, **kwargs):
        pass

    def _is_master_running(self):
        return True


class MockResolver:
    def __init__(self, *args, **kwargs):
        pass

    def mk_token(self, load):
        return {
            "token": "6d1b722e",
            "start": 10000.0,
            "expire": 20000.0,
            "name": "fred",
            "eauth": "rest",
            "auth_list": [
                "@test123",
            ],
        }

    def get_token(self, token):
        pass


@pytest.fixture
def configure_loader_modules():
    return {cherrypy_app: {}}


def test__loigin_rest_match_token():
    with patch("salt.netapi.rest_cherrypy.app.cherrypy", MockCherryPy()):
        with patch("salt.netapi.NetapiClient", MockNetapiClient):
            with patch("salt.auth.Resolver", MockResolver):
                login = cherrypy_app.Login()
                authtoken = login.POST()["return"][0]
                assert authtoken["token"] == "6d1b722e"


def test__login_rest_returns_perms():
    with patch("salt.netapi.rest_cherrypy.app.cherrypy", MockCherryPy()):
        with patch("salt.netapi.NetapiClient", MockNetapiClient):
            with patch("salt.auth.Resolver", MockResolver):
                login = cherrypy_app.Login()
                authtoken = login.POST()["return"][0]
                assert authtoken["perms"] == ["@test123"]



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    runtests_helpers.py
    ~~~~~~~~~~~~~~~~~~~
"""


import logging
import os
import sys
import tempfile

import salt.utils.platform

try:
    from tests.support.runtests import RUNTIME_VARS
except ImportError:
    # Salt SSH Tests
    SYS_TMP_DIR = os.path.realpath(
        # Avoid ${TMPDIR} and gettempdir() on MacOS as they yield a base path too long
        # for unix sockets: ``error: AF_UNIX path too long``
        # Gentoo Portage prefers ebuild tests are rooted in ${TMPDIR}
        os.environ.get("TMPDIR", tempfile.gettempdir())
        if not salt.utils.platform.is_darwin()
        else "/tmp"
    )
    # This tempdir path is defined on tests.integration.__init__
    TMP = os.path.join(SYS_TMP_DIR, "salt-tests-tmpdir")

    class RUNTIME_VARS:
        TMP = TMP
        SYS_TMP_DIR = SYS_TMP_DIR


log = logging.getLogger(__name__)


def get_salt_temp_dir():
    return RUNTIME_VARS.TMP


def get_salt_temp_dir_for_path(*path):
    return os.path.join(RUNTIME_VARS.TMP, *path)


def get_sys_temp_dir_for_path(*path):
    return os.path.join(RUNTIME_VARS.SYS_TMP_DIR, *path)


def nonzero_retcode_return_true():
    """
    Sets a nonzero retcode before returning. Designed to test orchestration.
    """
    __context__["retcode"] = 1
    return True


def nonzero_retcode_return_false():
    """
    Sets a nonzero retcode before returning. Designed to test orchestration.
    """
    __context__["retcode"] = 1
    return False


def fail_function(*args, **kwargs):  # pylint: disable=unused-argument
    """
    Return False no matter what is passed to it
    """
    return False


def get_python_executable():
    """
    Return the path to the python executable.

    This is particularly important when running the test suite within a virtualenv, while trying
    to create virtualenvs on windows.
    """
    try:
        if salt.utils.platform.is_windows():
            python_binary = os.path.join(
                sys.real_prefix, os.path.basename(sys.executable)
            )
        else:
            python_binary = os.path.join(
                sys.real_prefix, "bin", os.path.basename(sys.executable)
            )
            if not os.path.exists(python_binary):
                if not python_binary[-1].isdigit():
                    versioned_python_binary = "{}{}".format(
                        python_binary, *sys.version_info
                    )
                    log.info(
                        "Python binary could not be found at %s. Trying %s",
                        python_binary,
                        versioned_python_binary,
                    )
                    if os.path.exists(versioned_python_binary):
                        python_binary = versioned_python_binary
        if not os.path.exists(python_binary):
            log.warning("Python binary could not be found at %s", python_binary)
            python_binary = None
    except AttributeError:
        # We're not running inside a virtualenv
        python_binary = sys.executable
    return python_binary



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.log
    ~~~~~~~~

    This is where Salt's logging gets set up. Currently, the required imports
    are made to assure backwards compatibility.
"""

# pylint: disable = no-name-in-module

# Import several classes/functions from salt.log.setup for backwards compatibility
from salt._logging import LOG_LEVELS, SORTED_LEVEL_NAMES
from salt.log.setup import (
    is_console_configured,
    is_logfile_configured,
    is_logging_configured,
    is_temp_logging_configured,
    set_logger_level,
    setup_console_logger,
    setup_logfile_logger,
    setup_temp_logger,
)
from salt.utils.versions import warn_until_date

warn_until_date(
    "20240101",
    "Please stop using '{name}' and instead use 'salt._logging'. "
    "'{name}' will go away after {{date}}.".format(name=__name__),
    stacklevel=3,
)



"""
Module for using the locate utilities
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return (
            False,
            "The locate execution module cannot be loaded: only available on "
            "non-Windows systems.",
        )
    return True


def version():
    """
    Returns the version of locate

    CLI Example:

    .. code-block:: bash

        salt '*' locate.version
    """
    cmd = "locate -V"
    out = __salt__["cmd.run"](cmd).splitlines()
    return out


def stats():
    """
    Returns statistics about the locate database

    CLI Example:

    .. code-block:: bash

        salt '*' locate.stats
    """
    ret = {}
    cmd = "locate -S"
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        comps = line.strip().split()
        if line.startswith("Database"):
            ret["database"] = comps[1].replace(":", "")
            continue
        ret[" ".join(comps[1:])] = comps[0]
    return ret


def updatedb():
    """
    Updates the locate database

    CLI Example:

    .. code-block:: bash

        salt '*' locate.updatedb
    """
    cmd = "updatedb"
    out = __salt__["cmd.run"](cmd).splitlines()
    return out


def locate(pattern, database="", limit=0, **kwargs):
    """
    Performs a file lookup. Valid options (and their defaults) are::

        basename=False
        count=False
        existing=False
        follow=True
        ignore=False
        nofollow=False
        wholename=True
        regex=False
        database=<locate's default database>
        limit=<integer, not set by default>

    See the manpage for ``locate(1)`` for further explanation of these options.

    CLI Example:

    .. code-block:: bash

        salt '*' locate.locate
    """
    options = ""
    toggles = {
        "basename": "b",
        "count": "c",
        "existing": "e",
        "follow": "L",
        "ignore": "i",
        "nofollow": "P",
        "wholename": "w",
    }
    for option in kwargs:
        if bool(kwargs[option]) is True and option in toggles:
            options += toggles[option]
    if options:
        options = "-{}".format(options)
    if database:
        options += " -d {}".format(database)
    if limit > 0:
        options += " -l {}".format(limit)
    if "regex" in kwargs and bool(kwargs["regex"]) is True:
        options += " --regex"
    cmd = "locate {} {}".format(options, pattern)
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    return out



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.ldapmod
"""


import time

import pytest

import salt.modules.ldapmod as ldapmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {ldapmod: {}}


# 'search' function tests: 1


def test_search():
    """
    Test if it run an arbitrary LDAP query and return the results.
    """

    class MockConnect:
        """
        Mocking _connect method
        """

        def __init__(self):
            self.bdn = None
            self.scope = None
            self._filter = None
            self.attrs = None

        def search_s(self, bdn, scope, _filter, attrs):
            """
            Mock function for search_s
            """
            self.bdn = bdn
            self.scope = scope
            self._filter = _filter
            self.attrs = attrs
            return "SALT"

    mock = MagicMock(return_value=True)
    with patch.dict(ldapmod.__salt__, {"config.option": mock}):
        with patch.object(ldapmod, "_connect", MagicMock(return_value=MockConnect())):
            with patch.object(time, "time", MagicMock(return_value=8e-04)):
                assert ldapmod.search(filter="myhost") == {
                    "count": 4,
                    "results": "SALT",
                    "time": {"raw": "0.0", "human": "0.0ms"},
                }



import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
]

IS_WINDOWS = salt.utils.platform.is_windows()


def test_absent_file(file, tmp_path):
    """
    file.absent
    """
    name = tmp_path / "file_to_kill"
    name.write_text("killme")
    ret = file.absent(name=str(name))
    assert ret.result is True
    assert not name.is_file()


def test_absent_dir(file, tmp_path):
    """
    file.absent
    """
    name = tmp_path / "dir_to_kill"
    name.mkdir(exist_ok=True)
    ret = file.absent(name=str(name))
    assert ret.result is True
    assert not name.is_dir()


def test_absent_link(file, tmp_path):
    """
    file.absent
    """
    name = tmp_path / "link_to_kill"
    tgt = tmp_path / "link_to_kill.tgt"

    tgt.symlink_to(name, target_is_directory=IS_WINDOWS)

    ret = file.absent(name=str(name))

    assert ret.result is True
    assert not name.exists()
    assert not name.is_symlink()


def test_test_absent(file, tmp_path):
    """
    file.absent test interface
    """
    name = tmp_path / "testfile"
    name.write_text("killme")
    ret = file.absent(test=True, name=str(name))
    assert ret.result is None
    assert name.is_file()



"""
Test case for the consul state module
"""


import logging

import pytest

import salt.states.consul as consul
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        consul: {
            "__opts__": {
                "consul": {"url": "http://127.0.0.1", "token": "test_token"},
                "test": False,
            },
            "__grains__": {"id": "test-minion"},
        }
    }


def test_acl_present():
    """
    Test salt.states.consul.acl_present function
    """
    acl_info = {
        "data": [
            {
                "ID": "1d53dcd6-8f4f-431e-818a-9987996701a1",
                "Name": "89530557-8f18-4e29-a2d6-5b2fc8bca713",
                "Type": "client",
                "Rules": "",
                "CreateIndex": 419,
                "ModifyIndex": 420,
            }
        ],
        "res": True,
    }

    acl_info_mock = MagicMock(return_value=acl_info)
    with patch.dict(consul.__salt__, {"consul.acl_info": acl_info_mock}):
        with patch.object(consul, "_acl_changes", MagicMock(return_value=False)):
            consul_return = consul.acl_present(
                "my_acl",
                id="1d53dcd6-8f4f-431e-818a-9987996701a1",
                token="89530557-8f18-4e29-a2d6-5b2fc8bca713",
                type="client",
                consul_url="http://localhost:8500",
            )

            _expected = {
                "changes": {},
                "comment": 'ACL "my_acl" exists and is up to date',
                "name": "my_acl",
                "result": True,
            }
            assert consul_return == _expected

        acl_update_mock = MagicMock(
            return_value={
                "data": {"ID": "1D53DCD6-8F4F-431E-818A-9987996701A1"},
                "res": True,
            }
        )
        with patch.object(consul, "_acl_changes", MagicMock(return_value=True)):
            with patch.dict(consul.__salt__, {"consul.acl_update": acl_update_mock}):
                consul_return = consul.acl_present(
                    "my_acl",
                    id="1d53dcd6-8f4f-431e-818a-9987996701a1",
                    token="89530557-8f18-4e29-a2d6-5b2fc8bca713",
                    type="client",
                    consul_url="http://localhost:8500",
                )

            _expected = {
                "changes": {},
                "comment": "The acl has been updated",
                "name": "my_acl",
                "result": True,
            }
            assert consul_return == _expected


def test_acl_absent():
    """
    Test salt.states.consul.acl_absent function
    """
    #
    # Test when the ACL does exist
    #
    acl_info = {
        "data": [
            {
                "ID": "1d53dcd6-8f4f-431e-818a-9987996701a1",
                "Name": "89530557-8f18-4e29-a2d6-5b2fc8bca713",
                "Type": "client",
                "Rules": "",
                "CreateIndex": 419,
                "ModifyIndex": 420,
            }
        ],
        "res": True,
    }
    acl_info_mock = MagicMock(return_value=acl_info)
    acl_delete_mock = MagicMock(
        return_value={
            "res": True,
            "message": "ACL 38AC8470-4A83-4140-8DFD-F924CD32917F deleted.",
        }
    )
    with patch.dict(consul.__salt__, {"consul.acl_info": acl_info_mock}):
        with patch.dict(consul.__salt__, {"consul.acl_delete": acl_delete_mock}):
            consul_return = consul.acl_absent(
                "my_acl",
                id="1d53dcd6-8f4f-431e-818a-9987996701a1",
                token="89530557-8f18-4e29-a2d6-5b2fc8bca713",
                consul_url="http://localhost:8500",
            )

            _expected = {
                "changes": {},
                "comment": "The acl has been deleted",
                "name": "1d53dcd6-8f4f-431e-818a-9987996701a1",
                "result": True,
            }
            assert consul_return == _expected

    #
    # Test when the ACL does not exist
    #
    acl_info = {"data": [], "res": True}
    acl_info_mock = MagicMock(return_value=acl_info)
    acl_delete_mock = MagicMock(
        return_value={
            "res": True,
            "message": "ACL 38AC8470-4A83-4140-8DFD-F924CD32917F deleted.",
        }
    )
    with patch.dict(consul.__salt__, {"consul.acl_info": acl_info_mock}):
        with patch.dict(consul.__salt__, {"consul.acl_delete": acl_delete_mock}):
            consul_return = consul.acl_absent(
                "my_acl",
                id="1d53dcd6-8f4f-431e-818a-9987996701a1",
                token="89530557-8f18-4e29-a2d6-5b2fc8bca713",
                consul_url="http://localhost:8500",
            )

            _expected = {
                "changes": {},
                "comment": 'ACL "1d53dcd6-8f4f-431e-818a-9987996701a1" does not exist',
                "name": "1d53dcd6-8f4f-431e-818a-9987996701a1",
                "result": True,
            }
            assert consul_return == _expected



"""
Zeromq transport classes
"""
import errno
import hashlib
import logging
import os
import signal
import sys
import threading
from random import randint

import zmq.error
import zmq.eventloop.zmqstream

import salt.ext.tornado
import salt.ext.tornado.concurrent
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.ext.tornado.locks
import salt.payload
import salt.transport.base
import salt.utils.files
import salt.utils.process
import salt.utils.stringutils
import salt.utils.zeromq
from salt._compat import ipaddress
from salt.exceptions import SaltException, SaltReqTimeoutError
from salt.utils.zeromq import LIBZMQ_VERSION_INFO, ZMQ_VERSION_INFO, zmq

try:
    import zmq.utils.monitor

    HAS_ZMQ_MONITOR = True
except ImportError:
    HAS_ZMQ_MONITOR = False


log = logging.getLogger(__name__)


def _get_master_uri(master_ip, master_port, source_ip=None, source_port=None):
    """
    Return the ZeroMQ URI to connect the Minion to the Master.
    It supports different source IP / port, given the ZeroMQ syntax:
    // Connecting using a IP address and bind to an IP address
    rc = zmq_connect(socket, "tcp://192.168.1.17:5555;192.168.1.1:5555"); assert (rc == 0);
    Source: http://api.zeromq.org/4-1:zmq-tcp
    """
    from salt.utils.network import ip_bracket

    master_uri = "tcp://{master_ip}:{master_port}".format(
        master_ip=ip_bracket(master_ip), master_port=master_port
    )

    if source_ip or source_port:
        if LIBZMQ_VERSION_INFO >= (4, 1, 6) and ZMQ_VERSION_INFO >= (16, 0, 1):
            # The source:port syntax for ZeroMQ has been added in libzmq 4.1.6
            # which is included in the pyzmq wheels starting with 16.0.1.
            if source_ip and source_port:
                master_uri = (
                    "tcp://{source_ip}:{source_port};{master_ip}:{master_port}".format(
                        source_ip=ip_bracket(source_ip),
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
            elif source_ip and not source_port:
                master_uri = "tcp://{source_ip}:0;{master_ip}:{master_port}".format(
                    source_ip=ip_bracket(source_ip),
                    master_ip=ip_bracket(master_ip),
                    master_port=master_port,
                )
            elif source_port and not source_ip:
                ip_any = (
                    "0.0.0.0"
                    if ipaddress.ip_address(master_ip).version == 4
                    else ip_bracket("::")
                )
                master_uri = (
                    "tcp://{ip_any}:{source_port};{master_ip}:{master_port}".format(
                        ip_any=ip_any,
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
        else:
            log.warning(
                "Unable to connect to the Master using a specific source IP / port"
            )
            log.warning("Consider upgrading to pyzmq >= 16.0.1 and libzmq >= 4.1.6")
            log.warning(
                "Specific source IP / port for connecting to master returner port:"
                " configuraion ignored"
            )

    return master_uri


class PublishClient(salt.transport.base.PublishClient):
    """
    A transport channel backed by ZeroMQ for a Salt Publisher to use to
    publish commands to connected minions
    """

    ttype = "zeromq"

    def __init__(self, opts, io_loop, **kwargs):
        super().__init__(opts, io_loop, **kwargs)
        self.opts = opts
        self.io_loop = io_loop
        self.hexid = hashlib.sha1(
            salt.utils.stringutils.to_bytes(self.opts["id"])
        ).hexdigest()
        self._closing = False
        self.context = zmq.Context()
        self._socket = self.context.socket(zmq.SUB)
        if self.opts["zmq_filtering"]:
            # TODO: constants file for "broadcast"
            self._socket.setsockopt(zmq.SUBSCRIBE, b"broadcast")
            if self.opts.get("__role") == "syndic":
                self._socket.setsockopt(zmq.SUBSCRIBE, b"syndic")
            else:
                self._socket.setsockopt(
                    zmq.SUBSCRIBE, salt.utils.stringutils.to_bytes(self.hexid)
                )
        else:
            self._socket.setsockopt(zmq.SUBSCRIBE, b"")

        self._socket.setsockopt(
            zmq.IDENTITY, salt.utils.stringutils.to_bytes(self.opts["id"])
        )

        # TODO: cleanup all the socket opts stuff
        if hasattr(zmq, "TCP_KEEPALIVE"):
            self._socket.setsockopt(zmq.TCP_KEEPALIVE, self.opts["tcp_keepalive"])
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_IDLE, self.opts["tcp_keepalive_idle"]
            )
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_CNT, self.opts["tcp_keepalive_cnt"]
            )
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_INTVL, self.opts["tcp_keepalive_intvl"]
            )

        recon_delay = self.opts["recon_default"]

        if self.opts["recon_randomize"]:
            recon_delay = randint(
                self.opts["recon_default"],
                self.opts["recon_default"] + self.opts["recon_max"],
            )

            log.debug(
                "Generated random reconnect delay between '%sms' and '%sms' (%s)",
                self.opts["recon_default"],
                self.opts["recon_default"] + self.opts["recon_max"],
                recon_delay,
            )

        log.debug("Setting zmq_reconnect_ivl to '%sms'", recon_delay)
        self._socket.setsockopt(zmq.RECONNECT_IVL, recon_delay)

        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            log.debug(
                "Setting zmq_reconnect_ivl_max to '%sms'",
                self.opts["recon_default"] + self.opts["recon_max"],
            )

            self._socket.setsockopt(zmq.RECONNECT_IVL_MAX, self.opts["recon_max"])

        if (self.opts["ipv6"] is True or ":" in self.opts["master_ip"]) and hasattr(
            zmq, "IPV4ONLY"
        ):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self._socket.setsockopt(zmq.IPV4ONLY, 0)

        if HAS_ZMQ_MONITOR and self.opts["zmq_monitor"]:
            self._monitor = ZeroMQSocketMonitor(self._socket)
            self._monitor.start_io_loop(self.io_loop)

    def close(self):
        if self._closing is True:
            return
        self._closing = True
        if hasattr(self, "_monitor") and self._monitor is not None:
            self._monitor.stop()
            self._monitor = None
        if hasattr(self, "_stream"):
            self._stream.close(0)
        elif hasattr(self, "_socket"):
            self._socket.close(0)
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()

    # pylint: enable=W1701
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # TODO: this is the time to see if we are connected, maybe use the req channel to guess?
    @salt.ext.tornado.gen.coroutine
    def connect(self, publish_port, connect_callback=None, disconnect_callback=None):
        self.publish_port = publish_port
        log.debug(
            "Connecting the Minion to the Master publish port, using the URI: %s",
            self.master_pub,
        )
        log.debug("%r connecting to %s", self, self.master_pub)
        self._socket.connect(self.master_pub)
        connect_callback(True)

    @property
    def master_pub(self):
        """
        Return the master publish port
        """
        return _get_master_uri(
            self.opts["master_ip"],
            self.publish_port,
            source_ip=self.opts.get("source_ip"),
            source_port=self.opts.get("source_publish_port"),
        )

    @salt.ext.tornado.gen.coroutine
    def _decode_messages(self, messages):
        """
        Take the zmq messages, decrypt/decode them into a payload

        :param list messages: A list of messages to be decoded
        """
        messages_len = len(messages)
        # if it was one message, then its old style
        if messages_len == 1:
            payload = salt.payload.loads(messages[0])
        # 2 includes a header which says who should do it
        elif messages_len == 2:
            message_target = salt.utils.stringutils.to_str(messages[0])
            if (
                self.opts.get("__role") != "syndic"
                and message_target not in ("broadcast", self.hexid)
            ) or (
                self.opts.get("__role") == "syndic"
                and message_target not in ("broadcast", "syndic")
            ):
                log.debug("Publish received for not this minion: %s", message_target)
                raise salt.ext.tornado.gen.Return(None)
            payload = salt.payload.loads(messages[1])
        else:
            raise Exception(
                "Invalid number of messages ({}) in zeromq pubmessage from master".format(
                    len(messages_len)
                )
            )
        # Yield control back to the caller. When the payload has been decoded, assign
        # the decoded payload to 'ret' and resume operation
        raise salt.ext.tornado.gen.Return(payload)

    @property
    def stream(self):
        """
        Return the current zmqstream, creating one if necessary
        """
        if not hasattr(self, "_stream"):
            self._stream = zmq.eventloop.zmqstream.ZMQStream(
                self._socket, io_loop=self.io_loop
            )
        return self._stream

    def on_recv(self, callback):
        """
        Register a callback for received messages (that we didn't initiate)

        :param func callback: A function which should be called when data is received
        """
        return self.stream.on_recv(callback)

    @salt.ext.tornado.gen.coroutine
    def send(self, msg):
        self.stream.send(msg, noblock=True)


class RequestServer(salt.transport.base.DaemonizedRequestServer):
    def __init__(self, opts):  # pylint: disable=W0231
        self.opts = opts
        self._closing = False
        self._monitor = None
        self._w_monitor = None

    def zmq_device(self):
        """
        Multiprocessing target for the zmq queue device
        """
        self.__setup_signals()
        context = zmq.Context(self.opts["worker_threads"])
        # Prepare the zeromq sockets
        self.uri = "tcp://{interface}:{ret_port}".format(**self.opts)
        self.clients = context.socket(zmq.ROUTER)
        self.clients.setsockopt(zmq.LINGER, -1)
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self.clients.setsockopt(zmq.IPV4ONLY, 0)
        self.clients.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        self._start_zmq_monitor()
        self.workers = context.socket(zmq.DEALER)
        self.workers.setsockopt(zmq.LINGER, -1)

        if self.opts["mworker_queue_niceness"] and not salt.utils.platform.is_windows():
            log.info(
                "setting mworker_queue niceness to %d",
                self.opts["mworker_queue_niceness"],
            )
            os.nice(self.opts["mworker_queue_niceness"])

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )

        log.info("Setting up the master communication server")
        log.info("ReqServer clients %s", self.uri)
        self.clients.bind(self.uri)
        log.info("ReqServer workers %s", self.w_uri)
        self.workers.bind(self.w_uri)
        if self.opts.get("ipc_mode", "") != "tcp":
            os.chmod(os.path.join(self.opts["sock_dir"], "workers.ipc"), 0o600)

        while True:
            if self.clients.closed or self.workers.closed:
                break
            try:
                zmq.device(zmq.QUEUE, self.clients, self.workers)
            except zmq.ZMQError as exc:
                if exc.errno == errno.EINTR:
                    continue
                raise
            except (KeyboardInterrupt, SystemExit):
                break
        context.term()

    def close(self):
        """
        Cleanly shutdown the router socket
        """
        if self._closing:
            return
        log.info("MWorkerQueue under PID %s is closing", os.getpid())
        self._closing = True
        if getattr(self, "_monitor", None) is not None:
            self._monitor.stop()
            self._monitor = None
        if getattr(self, "_w_monitor", None) is not None:
            self._w_monitor.stop()
            self._w_monitor = None
        if hasattr(self, "clients") and self.clients.closed is False:
            self.clients.close()
        if hasattr(self, "workers") and self.workers.closed is False:
            self.workers.close()
        if hasattr(self, "stream"):
            self.stream.close()
        if hasattr(self, "_socket") and self._socket.closed is False:
            self._socket.close()
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()

    def pre_fork(self, process_manager):
        """
        Pre-fork we need to create the zmq router device

        :param func process_manager: An instance of salt.utils.process.ProcessManager
        """
        process_manager.add_process(self.zmq_device, name="MWorkerQueue")

    def _start_zmq_monitor(self):
        """
        Starts ZMQ monitor for debugging purposes.
        :return:
        """
        # Socket monitor shall be used the only for debug
        # purposes so using threading doesn't look too bad here

        if HAS_ZMQ_MONITOR and self.opts["zmq_monitor"]:
            log.debug("Starting ZMQ monitor")
            self._w_monitor = ZeroMQSocketMonitor(self._socket)
            threading.Thread(target=self._w_monitor.start_poll).start()
            log.debug("ZMQ monitor has been started started")

    def post_fork(self, message_handler, io_loop):
        """
        After forking we need to create all of the local sockets to listen to the
        router

        :param func message_handler: A function to called to handle incoming payloads as
                                     they are picked up off the wire
        :param IOLoop io_loop: An instance of a Tornado IOLoop, to handle event scheduling
        """
        context = zmq.Context(1)
        self._socket = context.socket(zmq.REP)
        # Linger -1 means we'll never discard messages.
        self._socket.setsockopt(zmq.LINGER, -1)
        self._start_zmq_monitor()

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )
        log.info("Worker binding to socket %s", self.w_uri)
        self._socket.connect(self.w_uri)
        if self.opts.get("ipc_mode", "") != "tcp" and os.path.isfile(
            os.path.join(self.opts["sock_dir"], "workers.ipc")
        ):
            os.chmod(os.path.join(self.opts["sock_dir"], "workers.ipc"), 0o600)
        self.stream = zmq.eventloop.zmqstream.ZMQStream(self._socket, io_loop=io_loop)
        self.message_handler = message_handler
        self.stream.on_recv_stream(self.handle_message)

    @salt.ext.tornado.gen.coroutine
    def handle_message(self, stream, payload):
        try:
            payload = self.decode_payload(payload)
        except salt.exceptions.SaltDeserializationError:
            self.stream.send(self.encode_payload({"msg": "bad load"}))
            return
        # XXX: Is header really needed?
        reply = yield self.message_handler(payload)
        self.stream.send(self.encode_payload(reply))

    def encode_payload(self, payload):
        return salt.payload.dumps(payload)

    def __setup_signals(self):
        signal.signal(signal.SIGINT, self._handle_signals)
        signal.signal(signal.SIGTERM, self._handle_signals)

    def _handle_signals(self, signum, sigframe):
        msg = "{} received a ".format(self.__class__.__name__)
        if signum == signal.SIGINT:
            msg += "SIGINT"
        elif signum == signal.SIGTERM:
            msg += "SIGTERM"
        msg += ". Exiting"
        log.debug(msg)
        self.close()
        sys.exit(salt.defaults.exitcodes.EX_OK)

    def decode_payload(self, payload):
        payload = salt.payload.loads(payload[0])
        return payload


def _set_tcp_keepalive(zmq_socket, opts):
    """
    Ensure that TCP keepalives are set as specified in "opts".

    Warning: Failure to set TCP keepalives on the salt-master can result in
    not detecting the loss of a minion when the connection is lost or when
    its host has been terminated without first closing the socket.
    Salt's Presence System depends on this connection status to know if a minion
    is "present".

    Warning: Failure to set TCP keepalives on minions can result in frequent or
    unexpected disconnects!
    """
    if hasattr(zmq, "TCP_KEEPALIVE") and opts:
        if "tcp_keepalive" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE, opts["tcp_keepalive"])
        if "tcp_keepalive_idle" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, opts["tcp_keepalive_idle"])
        if "tcp_keepalive_cnt" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_CNT, opts["tcp_keepalive_cnt"])
        if "tcp_keepalive_intvl" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, opts["tcp_keepalive_intvl"])


# TODO: unit tests!
class AsyncReqMessageClient:
    """
    This class wraps the underlying zeromq REQ socket and gives a future-based
    interface to sending and recieving messages. This works around the primary
    limitation of serialized send/recv on the underlying socket by queueing the
    message sends in this class. In the future if we decide to attempt to multiplex
    we can manage a pool of REQ/REP sockets-- but for now we'll just do them in serial
    """

    def __init__(self, opts, addr, linger=0, io_loop=None):
        """
        Create an asynchronous message client

        :param dict opts: The salt opts dictionary
        :param str addr: The interface IP address to bind to
        :param int linger: The number of seconds to linger on a ZMQ socket. See
                           http://api.zeromq.org/2-1:zmq-setsockopt [ZMQ_LINGER]
        :param IOLoop io_loop: A Tornado IOLoop event scheduler [tornado.ioloop.IOLoop]
        """
        self.opts = opts
        self.addr = addr
        self.linger = linger
        if io_loop is None:
            self.io_loop = salt.ext.tornado.ioloop.IOLoop.current()
        else:
            self.io_loop = io_loop

        self.context = zmq.Context()

        self.send_queue = []

        self._closing = False
        self._future = None
        self.lock = salt.ext.tornado.locks.Lock()

    def connect(self):
        if hasattr(self, "stream"):
            return
        # wire up sockets
        self._init_socket()

    # TODO: timeout all in-flight sessions, or error
    def close(self):
        try:
            if self._closing:
                return
        except AttributeError:
            # We must have been called from __del__
            # The python interpreter has nuked most attributes already
            return
        else:
            self._closing = True
            if hasattr(self, "stream") and self.stream is not None:
                if ZMQ_VERSION_INFO < (14, 3, 0):
                    # stream.close() doesn't work properly on pyzmq < 14.3.0
                    if self.stream.socket:
                        self.stream.socket.close()
                    self.stream.io_loop.remove_handler(self.stream.socket)
                    # set this to None, more hacks for messed up pyzmq
                    self.stream.socket = None
                    self.socket.close()
                else:
                    self.stream.close(1)
                    self.socket = None
                self.stream = None
            if self._future:
                self._future.set_exception(SaltException("Closing connection"))
                self._future = None
            if self.context.closed is False:
                # This hangs if closing the stream causes an import error
                self.context.term()

    def _init_socket(self):
        self.socket = self.context.socket(zmq.REQ)

        # socket options
        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            self.socket.setsockopt(zmq.RECONNECT_IVL_MAX, 5000)

        _set_tcp_keepalive(self.socket, self.opts)
        if self.addr.startswith("tcp://["):
            # Hint PF type if bracket enclosed IPv6 address
            if hasattr(zmq, "IPV6"):
                self.socket.setsockopt(zmq.IPV6, 1)
            elif hasattr(zmq, "IPV4ONLY"):
                self.socket.setsockopt(zmq.IPV4ONLY, 0)
        self.socket.linger = self.linger
        self.socket.connect(self.addr)
        self.stream = zmq.eventloop.zmqstream.ZMQStream(
            self.socket, io_loop=self.io_loop
        )
        self.stream.on_recv(self.handle_reply)

    def timeout_message(self, future):
        """
        Handle a message timeout by removing it from the sending queue
        and informing the caller

        :raises: SaltReqTimeoutError
        """
        if self._future == future:
            self._future = None
            future.set_exception(SaltReqTimeoutError("Message timed out"))

    @salt.ext.tornado.gen.coroutine
    def send(self, message, timeout=None, callback=None):
        """
        Return a future which will be completed when the message has a response
        """
        future = salt.ext.tornado.concurrent.Future()

        message = salt.payload.dumps(message)

        if callback is not None:

            def handle_future(future):
                response = future.result()
                self.io_loop.add_callback(callback, response)

            future.add_done_callback(handle_future)

        if self.opts.get("detect_mode") is True:
            timeout = 1

        if timeout is not None:
            send_timeout = self.io_loop.call_later(
                timeout, self.timeout_message, future
            )

        with (yield self.lock.acquire()):
            self._future = future
            yield self.stream.send(message)
            recv = yield future
        raise salt.ext.tornado.gen.Return(recv)

    def handle_reply(self, msg):
        data = salt.payload.loads(msg[0])
        future = self._future
        self._future = None
        future.set_result(data)


class ZeroMQSocketMonitor:
    __EVENT_MAP = None

    def __init__(self, socket):
        """
        Create ZMQ monitor sockets

        More information:
            http://api.zeromq.org/4-0:zmq-socket-monitor
        """
        self._socket = socket
        self._monitor_socket = self._socket.get_monitor_socket()
        self._monitor_stream = None

    def start_io_loop(self, io_loop):
        log.trace("Event monitor start!")
        self._monitor_stream = zmq.eventloop.zmqstream.ZMQStream(
            self._monitor_socket, io_loop=io_loop
        )
        self._monitor_stream.on_recv(self.monitor_callback)

    def start_poll(self):
        log.trace("Event monitor start!")
        try:
            while self._monitor_socket is not None and self._monitor_socket.poll():
                msg = self._monitor_socket.recv_multipart()
                self.monitor_callback(msg)
        except (AttributeError, zmq.error.ContextTerminated):
            # We cannot log here because we'll get an interrupted system call in trying
            # to flush the logging buffer as we terminate
            pass

    @property
    def event_map(self):
        if ZeroMQSocketMonitor.__EVENT_MAP is None:
            event_map = {}
            for name in dir(zmq):
                if name.startswith("EVENT_"):
                    value = getattr(zmq, name)
                    event_map[value] = name
            ZeroMQSocketMonitor.__EVENT_MAP = event_map
        return ZeroMQSocketMonitor.__EVENT_MAP

    def monitor_callback(self, msg):
        evt = zmq.utils.monitor.parse_monitor_message(msg)
        evt["description"] = self.event_map[evt["event"]]
        log.debug("ZeroMQ event: %s", evt)
        if evt["event"] == zmq.EVENT_MONITOR_STOPPED:
            self.stop()

    def stop(self):
        if self._socket is None:
            return
        self._socket.disable_monitor()
        self._socket = None
        self._monitor_socket = None
        if self._monitor_stream is not None:
            self._monitor_stream.close()
            self._monitor_stream = None
        log.trace("Event monitor done!")


class PublishServer(salt.transport.base.DaemonizedPublishServer):
    """
    Encapsulate synchronous operations for a publisher channel
    """

    _sock_data = threading.local()

    def __init__(self, opts):
        self.opts = opts

    def connect(self):
        return salt.ext.tornado.gen.sleep(5)

    def publish_daemon(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
    ):
        """
        This method represents the Publish Daemon process. It is intended to be
        run in a thread or process as it creates and runs an it's own ioloop.
        """
        ioloop = salt.ext.tornado.ioloop.IOLoop()
        ioloop.make_current()
        self.io_loop = ioloop
        context = zmq.Context(1)
        pub_sock = context.socket(zmq.PUB)
        monitor = ZeroMQSocketMonitor(pub_sock)
        monitor.start_io_loop(ioloop)
        _set_tcp_keepalive(pub_sock, self.opts)
        self.dpub_sock = pub_sock = zmq.eventloop.zmqstream.ZMQStream(pub_sock)
        # if 2.1 >= zmq < 3.0, we only have one HWM setting
        try:
            pub_sock.setsockopt(zmq.HWM, self.opts.get("pub_hwm", 1000))
        # in zmq >= 3.0, there are separate send and receive HWM settings
        except (AttributeError, zmq.error.ZMQError):
            # Set the High Water Marks. For more information on HWM, see:
            # http://api.zeromq.org/4-1:zmq-setsockopt
            pub_sock.setsockopt(zmq.SNDHWM, self.opts.get("pub_hwm", 1000))
            pub_sock.setsockopt(zmq.RCVHWM, self.opts.get("pub_hwm", 1000))
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            pub_sock.setsockopt(zmq.IPV4ONLY, 0)
        pub_sock.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        pub_sock.setsockopt(zmq.LINGER, -1)
        # Prepare minion pull socket
        pull_sock = context.socket(zmq.PULL)
        pull_sock = zmq.eventloop.zmqstream.ZMQStream(pull_sock)
        pull_sock.setsockopt(zmq.LINGER, -1)
        salt.utils.zeromq.check_ipc_path_max_len(self.pull_uri)
        # Start the minion command publisher
        log.info("Starting the Salt Publisher on %s", self.pub_uri)
        pub_sock.bind(self.pub_uri)
        # Securely create socket
        log.info("Starting the Salt Puller on %s", self.pull_uri)
        with salt.utils.files.set_umask(0o177):
            pull_sock.bind(self.pull_uri)

        @salt.ext.tornado.gen.coroutine
        def on_recv(packages):
            for package in packages:
                payload = salt.payload.loads(package)
                yield publish_payload(payload)

        pull_sock.on_recv(on_recv)
        try:
            ioloop.start()
        finally:
            pub_sock.close()
            pull_sock.close()

    @property
    def pull_uri(self):
        if self.opts.get("ipc_mode", "") == "tcp":
            pull_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_publish_pull", 4514)
            )
        else:
            pull_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
            )
        return pull_uri

    @property
    def pub_uri(self):
        return "tcp://{interface}:{publish_port}".format(**self.opts)

    @salt.ext.tornado.gen.coroutine
    def publish_payload(self, payload, topic_list=None):
        payload = salt.payload.dumps(payload)
        if self.opts["zmq_filtering"]:
            if topic_list:
                for topic in topic_list:
                    log.trace("Sending filtered data over publisher %s", self.pub_uri)
                    # zmq filters are substring match, hash the topic
                    # to avoid collisions
                    htopic = salt.utils.stringutils.to_bytes(
                        hashlib.sha1(salt.utils.stringutils.to_bytes(topic)).hexdigest()
                    )
                    yield self.dpub_sock.send(htopic, flags=zmq.SNDMORE)
                    yield self.dpub_sock.send(payload)
                    log.trace("Filtered data has been sent")
                # Syndic broadcast
                if self.opts.get("order_masters"):
                    log.trace("Sending filtered data to syndic")
                    yield self.dpub_sock.send(b"syndic", flags=zmq.SNDMORE)
                    yield self.dpub_sock.send(payload)
                    log.trace("Filtered data has been sent to syndic")
            # otherwise its a broadcast
            else:
                # TODO: constants file for "broadcast"
                log.trace("Sending broadcasted data over publisher %s", self.pub_uri)
                yield self.dpub_sock.send(b"broadcast", flags=zmq.SNDMORE)
                yield self.dpub_sock.send(payload)
                log.trace("Broadcasted data has been sent")
        else:
            log.trace("Sending ZMQ-unfiltered data over publisher %s", self.pub_uri)
            yield self.dpub_sock.send(payload)
            log.trace("Unfiltered data has been sent")

    def pre_fork(self, process_manager):
        """
        Do anything necessary pre-fork. Since this is on the master side this will
        primarily be used to create IPC channels and create our daemon process to
        do the actual publishing

        :param func process_manager: A ProcessManager, from salt.utils.process.ProcessManager
        """
        process_manager.add_process(
            self.publish_daemon,
            args=(self.publish_payload,),
        )

    @property
    def pub_sock(self):
        """
        This thread's zmq publisher socket. This socket is stored on the class
        so that multiple instantiations in the same thread will re-use a single
        zmq socket.
        """
        try:
            return self._sock_data.sock
        except AttributeError:
            pass

    def pub_connect(self):
        """
        Create and connect this thread's zmq socket. If a publisher socket
        already exists "pub_close" is called before creating and connecting a
        new socket.
        """
        if self.pub_sock:
            self.pub_close()
        ctx = zmq.Context()
        self._sock_data.sock = ctx.socket(zmq.PUSH)
        self.pub_sock.setsockopt(zmq.LINGER, -1)
        if self.opts.get("ipc_mode", "") == "tcp":
            pull_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_publish_pull", 4514)
            )
        else:
            pull_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
            )
        log.debug("Connecting to pub server: %s", pull_uri)
        self.pub_sock.connect(pull_uri)
        return self._sock_data.sock

    def pub_close(self):
        """
        Disconnect an existing publisher socket and remove it from the local
        thread's cache.
        """
        if hasattr(self._sock_data, "sock"):
            self._sock_data.sock.close()
            delattr(self._sock_data, "sock")

    def publish(self, payload, **kwargs):
        """
        Publish "load" to minions. This send the load to the publisher daemon
        process with does the actual sending to minions.

        :param dict load: A load to be sent across the wire to minions
        """
        if not self.pub_sock:
            self.pub_connect()
        serialized = salt.payload.dumps(payload)
        self.pub_sock.send(serialized)
        log.debug("Sent payload to publish daemon.")

    @property
    def topic_support(self):
        return self.opts.get("zmq_filtering", False)

    def close(self):
        self.pub_close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class RequestClient(salt.transport.base.RequestClient):

    ttype = "zeromq"

    def __init__(self, opts, io_loop):  # pylint: disable=W0231
        self.opts = opts
        master_uri = self.get_master_uri(opts)
        self.message_client = AsyncReqMessageClient(
            self.opts,
            master_uri,
            io_loop=io_loop,
        )

    def connect(self):
        self.message_client.connect()

    @salt.ext.tornado.gen.coroutine
    def send(self, load, timeout=60):
        self.connect()
        ret = yield self.message_client.send(load, timeout=timeout)
        raise salt.ext.tornado.gen.Return(ret)

    def close(self):
        self.message_client.close()

    @staticmethod
    def get_master_uri(opts):
        if "master_uri" in opts:
            return opts["master_uri"]
        if "master_ip" in opts:
            return _get_master_uri(
                opts["master_ip"],
                opts["master_port"],
                source_ip=opts.get("source_ip"),
                source_port=opts.get("source_ret_port"),
            )
        # if we've reached here something is very abnormal
        raise SaltException("ReqChannel: missing master_uri/master_ip in self.opts")



import textwrap

import pytest

import salt.modules.pkgng as pkgng
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pkgng: {}}


@pytest.fixture
def pkgs():
    return [
        {"openvpn": "2.4.8_2"},
        {"openvpn": "2.4.8_2", "gettext-runtime": "0.20.1", "p5-Mojolicious": "8.40"},
    ]


def test_latest_version(pkgs):
    """
    Test basic usage of pkgng.latest_version
    """
    pkgs_mock = MagicMock(side_effect=pkgs)
    search_cmd = MagicMock(return_value="bash-5.1.4")
    with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
        with patch.dict(pkgng.__salt__, {"cmd.run": search_cmd}):
            result = pkgng.latest_version("bash")
            search_cmd.assert_called_with(
                ["pkg", "search", "-eqS", "name", "-U", "bash"],
                output_loglevel="trace",
                python_shell=False,
            )
            assert result == "5.1.4"


def test_latest_version_origin(pkgs):
    """
    Test pkgng.latest_version with a specific package origin
    """
    pkgs_mock = MagicMock(side_effect=pkgs)
    search_cmd = MagicMock(return_value="bash-5.1.4_2")
    with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
        with patch.dict(pkgng.__salt__, {"cmd.run": search_cmd}):
            result = pkgng.latest_version("shells/bash")
            search_cmd.assert_called_with(
                ["pkg", "search", "-eqS", "origin", "-U", "shells/bash"],
                output_loglevel="trace",
                python_shell=False,
            )
            assert result == "5.1.4_2"


def test_latest_version_outofdatedate(pkgs):
    """
    Test pkgng.latest_version with an out-of-date package
    """
    pkgs_mock = MagicMock(side_effect=pkgs)
    search_cmd = MagicMock(return_value="openvpn-2.4.8_3")
    with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
        with patch.dict(pkgng.__salt__, {"cmd.run": search_cmd}):
            result = pkgng.latest_version("openvpn")
            search_cmd.assert_called_with(
                ["pkg", "search", "-eqS", "name", "-U", "openvpn"],
                output_loglevel="trace",
                python_shell=False,
            )
            assert result == "2.4.8_3"


def test_latest_version_unavailable(pkgs):
    """
    Test pkgng.latest_version when the requested package is not available
    """
    pkgs_mock = MagicMock(side_effect=pkgs)
    search_cmd = MagicMock(return_value="")
    with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
        with patch.dict(pkgng.__salt__, {"cmd.run": search_cmd}):
            result = pkgng.latest_version("does_not_exist")
            search_cmd.assert_called_with(
                ["pkg", "search", "-eqS", "name", "-U", "does_not_exist"],
                output_loglevel="trace",
                python_shell=False,
            )


def test_latest_version_uptodate(pkgs):
    """
    Test pkgng.latest_version with an up-to-date package
    """
    pkgs_mock = MagicMock(side_effect=pkgs)
    search_cmd = MagicMock(return_value="openvpn-2.4.8_2")
    with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
        with patch.dict(pkgng.__salt__, {"cmd.run": search_cmd}):
            result = pkgng.latest_version("openvpn")
            search_cmd.assert_called_with(
                ["pkg", "search", "-eqS", "name", "-U", "openvpn"],
                output_loglevel="trace",
                python_shell=False,
            )
            assert result == ""


def test_lock():
    """
    Test pkgng.lock
    """
    lock_cmd = MagicMock(return_value={"stdout": "pkga-1.0\npkgb-2.0\n", "retcode": 0})
    with patch.dict(pkgng.__salt__, {"cmd.run_all": lock_cmd}):

        result = pkgng.lock("pkga")
        assert result
        lock_cmd.assert_called_with(
            ["pkg", "lock", "-y", "--quiet", "--show-locked", "pkga"],
            output_loglevel="trace",
            python_shell=False,
        )

        result = pkgng.lock("dummy")
        assert not result
        lock_cmd.assert_called_with(
            ["pkg", "lock", "-y", "--quiet", "--show-locked", "dummy"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_unlock():
    """
    Test pkgng.unlock
    """
    unlock_cmd = MagicMock(
        return_value={"stdout": "pkga-1.0\npkgb-2.0\n", "retcode": 0}
    )
    with patch.dict(pkgng.__salt__, {"cmd.run_all": unlock_cmd}):

        result = pkgng.unlock("pkga")
        assert not result
        unlock_cmd.assert_called_with(
            ["pkg", "unlock", "-y", "--quiet", "--show-locked", "pkga"],
            output_loglevel="trace",
            python_shell=False,
        )

        result = pkgng.unlock("dummy")
        assert result
        unlock_cmd.assert_called_with(
            ["pkg", "unlock", "-y", "--quiet", "--show-locked", "dummy"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_locked():
    """
    Test pkgng.unlock
    """
    lock_cmd = MagicMock(return_value={"stdout": "pkga-1.0\npkgb-2.0\n", "retcode": 0})
    with patch.dict(pkgng.__salt__, {"cmd.run_all": lock_cmd}):

        result = pkgng.locked("pkga")
        assert result
        lock_cmd.assert_called_with(
            ["pkg", "lock", "-y", "--quiet", "--show-locked"],
            output_loglevel="trace",
            python_shell=False,
        )

        result = pkgng.locked("dummy")
        assert not result
        lock_cmd.assert_called_with(
            ["pkg", "lock", "-y", "--quiet", "--show-locked"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_list_upgrades_present():
    """
    Test pkgng.list_upgrades with upgrades available
    """
    pkg_cmd = MagicMock(
        return_value=textwrap.dedent(
            """
        The following 6 package(s) will be affected (of 0 checked):

        Installed packages to be UPGRADED:
                pkga: 1.0 -> 1.1
                pkgb: 2.0 -> 2.1 [FreeBSD]
                pkgc: 3.0 -> 3.1 [FreeBSD] (dependency changed)
                pkgd: 4.0 -> 4.1 (dependency changed)

        New packages to be INSTALLED:
                pkge: 1.0
                pkgf: 2.0 [FreeBSD]
                pkgg: 3.0 [FreeBSD] (dependency changed)
                pkgh: 4.0 (dependency changed)

        Installed packages to be REINSTALLED:
                pkgi-1.0
                pkgj-2.0 [FreeBSD]
                pkgk-3.0 [FreeBSD] (direct dependency changed: pkga)
                pkgl-4.0 (direct dependency changed: pkgb)

        Installed packages to be DOWNGRADED:
                pkgm: 1.1 -> 1.0
                pkgn: 2.1 -> 2.0 [FreeBSD]
                pkgo: 3.1 -> 3.0 [FreeBSD] (dependency changed)
                pkgp: 4.1 -> 4.0 (dependency changed)

        Installed packages to be REMOVED:
                pkgq-1.0
                pkgr-2.0 [FreeBSD]
                pkgs-3.0 [FreeBSD] (direct dependency changed: pkga)
                pkgt-4.0 (direct dependency changed: pkgb)

        Number of packages to be upgraded: 2
        Number of packages to be reinstalled: 2

        The process will require 14 MiB more space.
        22 MiB to be downloaded.
        """
        )
    )

    with patch.dict(pkgng.__salt__, {"cmd.run_stdout": pkg_cmd}):

        result = pkgng.list_upgrades(refresh=False)
        assert result == {"pkga": "1.1", "pkgb": "2.1", "pkgc": "3.1", "pkgd": "4.1"}
        pkg_cmd.assert_called_with(
            ["pkg", "upgrade", "--dry-run", "--quiet", "--no-repo-update"],
            output_loglevel="trace",
            python_shell=False,
            ignore_retcode=True,
        )


def test_list_upgrades_absent():
    """
    Test pkgng.list_upgrades with no upgrades available
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run_stdout": pkg_cmd}):
        result = pkgng.list_upgrades(refresh=False)
        assert result == {}
        pkg_cmd.assert_called_with(
            ["pkg", "upgrade", "--dry-run", "--quiet", "--no-repo-update"],
            output_loglevel="trace",
            python_shell=False,
            ignore_retcode=True,
        )


def test_upgrade_without_fromrepo(pkgs):
    """
    Test pkg upgrade to upgrade all available packages
    """
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)

    with patch.dict(pkgng.__salt__, {"cmd.run_all": pkg_cmd}):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            result = pkgng.upgrade()
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert result == expected
            pkg_cmd.assert_called_with(
                ["pkg", "upgrade", "-y"],
                output_loglevel="trace",
                python_shell=False,
            )


def test_upgrade_with_fromrepo(pkgs):
    """
    Test pkg upgrade to upgrade all available packages
    """
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)

    with patch.dict(pkgng.__salt__, {"cmd.run_all": pkg_cmd}):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            result = pkgng.upgrade(fromrepo="FreeBSD")
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert result == expected
            pkg_cmd.assert_called_with(
                ["pkg", "upgrade", "-y", "--repository", "FreeBSD"],
                output_loglevel="trace",
                python_shell=False,
            )


def test_upgrade_with_fetchonly(pkgs):
    """
    Test pkg upgrade to fetch packages only
    """
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)

    with patch.dict(pkgng.__salt__, {"cmd.run_all": pkg_cmd}):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            result = pkgng.upgrade(fetchonly=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert result == expected
            pkg_cmd.assert_called_with(
                ["pkg", "upgrade", "-Fy"],
                output_loglevel="trace",
                python_shell=False,
            )


def test_upgrade_with_local(pkgs):
    """
    Test pkg upgrade to supress automatic update of the local copy of the
    repository catalogue from remote
    """
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)

    with patch.dict(pkgng.__salt__, {"cmd.run_all": pkg_cmd}):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            result = pkgng.upgrade(local=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert result == expected
            pkg_cmd.assert_called_with(
                ["pkg", "upgrade", "-Uy"],
                output_loglevel="trace",
                python_shell=False,
            )


def test_stats_with_local():
    """
    Test pkg.stats for local packages
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.stats(local=True)
        assert result == []
        pkg_cmd.assert_called_with(
            ["pkg", "stats", "-l"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_stats_with_remote():
    """
    Test pkg.stats for remote packages
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.stats(remote=True)
        assert result == []
        pkg_cmd.assert_called_with(
            ["pkg", "stats", "-r"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_stats_with_bytes_remote():
    """
    Test pkg.stats to show disk space usage in bytes only for remote
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.stats(remote=True, bytes=True)
        assert result == []
        pkg_cmd.assert_called_with(
            ["pkg", "stats", "-rb"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_stats_with_bytes_local():
    """
    Test pkg.stats to show disk space usage in bytes only for local
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.stats(local=True, bytes=True)
        assert result == []
        pkg_cmd.assert_called_with(
            ["pkg", "stats", "-lb"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_install_without_args(pkgs):
    """
    Test pkg.install to install a package without arguments
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install()
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-y", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_local(pkgs):
    """
    Test pkg.install to install a package with local=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(local=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yU", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_fromrepo(pkgs):
    """
    Test pkg.install to install a package with fromrepo=FreeBSD argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(fromrepo="FreeBSD")
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                [
                    "pkg",
                    "install",
                    "-r",
                    "FreeBSD",
                    "-y",
                    "gettext-runtime",
                    "p5-Mojolicious",
                ],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_glob(pkgs):
    """
    Test pkg.install to install a package with glob=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(glob=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yg", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_reinstall_requires(pkgs):
    """
    Test pkg.install to install a package with reinstall_requires=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(reinstall_requires=True, force=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yfR", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_regex(pkgs):
    """
    Test pkg.install to install a package with regex=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(regex=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yx", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_batch(pkgs):
    """
    Test pkg.install to install a package with batch=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(batch=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-y", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={"BATCH": "true", "ASSUME_ALWAYS_YES": "YES"},
            )


def test_install_with_pcre(pkgs):
    """
    Test pkg.install to install a package with pcre=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(pcre=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yX", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_install_with_orphan(pkgs):
    """
    Test pkg.install to install a package with orphan=True argument
    """
    parsed_targets = (
        OrderedDict((("gettext-runtime", None), ("p5-Mojolicious", None))),
        "repository",
    )
    pkg_cmd = MagicMock(return_value={"retcode": 0})
    pkgs_mock = MagicMock(side_effect=pkgs)
    patches = {
        "cmd.run_all": pkg_cmd,
        "pkg_resource.parse_targets": MagicMock(return_value=parsed_targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        with patch("salt.modules.pkgng.list_pkgs", pkgs_mock):
            added = pkgng.install(orphan=True)
            expected = {
                "gettext-runtime": {"new": "0.20.1", "old": ""},
                "p5-Mojolicious": {"new": "8.40", "old": ""},
            }
            assert added == expected
            pkg_cmd.assert_called_with(
                ["pkg", "install", "-yA", "gettext-runtime", "p5-Mojolicious"],
                output_loglevel="trace",
                python_shell=False,
                env={},
            )


def test_check_depends():
    """
    Test pkgng.check to check and install missing dependencies
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.check(depends=True)
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "check", "-dy"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_check_checksum():
    """
    Test pkgng.check for packages with invalid checksums
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.check(checksum=True)
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "check", "-s"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_check_recompute():
    """
    Test pkgng.check to recalculate the checksums of installed packages
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.check(recompute=True)
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "check", "-r"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_check_checklibs():
    """
    Test pkgng.check to regenerate the library dependency metadata
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.check(checklibs=True)
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "check", "-B"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_autoremove_with_dryrun():
    """
    Test pkgng.autoremove with dryrun argument
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):
        result = pkgng.autoremove(dryrun=True)
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "autoremove", "-n"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_autoremove():
    """
    Test pkgng.autoremove
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):

        result = pkgng.autoremove()
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "autoremove", "-y"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_audit():
    """
    Test pkgng.audit
    """
    pkg_cmd = MagicMock(return_value="")

    with patch.dict(pkgng.__salt__, {"cmd.run": pkg_cmd}):

        result = pkgng.audit()
        assert result == ""
        pkg_cmd.assert_called_with(
            ["pkg", "audit", "-F"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_version():
    """
    Test pkgng.version
    """
    version = "2.0.6"
    mock = MagicMock(return_value=version)
    with patch.dict(pkgng.__salt__, {"pkg_resource.version": mock}):

        assert pkgng.version(*["mutt"]) == version
        assert not pkgng.version(*["mutt"]) == "2.0.10"


def test_refresh_db_without_forced_flag():
    """
    Test pkgng.refresh_db with force=False
    """
    pkg_cmd = MagicMock(return_value=0)
    with patch("salt.utils.pkg.clear_rtag", MagicMock()):
        with patch.dict(pkgng.__salt__, {"cmd.retcode": pkg_cmd}):

            result = pkgng.refresh_db()
            assert result is True
            pkg_cmd.assert_called_with(
                ["pkg", "update"],
                python_shell=False,
            )


def test_refresh_db_with_forced_flag():
    """
    Test pkgng.refresh_db with force=True
    """
    pkg_cmd = MagicMock(return_value=0)
    with patch("salt.utils.pkg.clear_rtag", MagicMock()):
        with patch.dict(pkgng.__salt__, {"cmd.retcode": pkg_cmd}):

            result = pkgng.refresh_db(force=True)
            assert result is True
            pkg_cmd.assert_called_with(
                ["pkg", "update", "-f"],
                python_shell=False,
            )


def test_fetch_with_default_flag():
    """
    Test pkgng.fetch with default options
    """

    targets = "mutt"
    pkg_cmd = MagicMock(return_value=targets)
    patches = {
        "cmd.run": pkg_cmd,
        "pkg_resource.stringify": MagicMock(),
        "pkg_resource.parse_targets": MagicMock(return_value=targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        pkgs = pkgng.fetch(targets)
        assert pkgs == targets
    pkg_cmd.assert_called_once_with(
        ["pkg", "fetch", "-y", "-g", "mutt"],
        output_loglevel="trace",
        python_shell=False,
    )


def test_fetch_with_dependency_flag():
    """
    Test pkgng.fetch with enabled dependency flag
    """

    targets = "mutt"
    pkg_cmd = MagicMock(return_value=targets)
    patches = {
        "cmd.run": pkg_cmd,
        "pkg_resource.stringify": MagicMock(),
        "pkg_resource.parse_targets": MagicMock(return_value=targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        pkgs = pkgng.fetch(targets, depends=True)
        assert pkgs == targets
    pkg_cmd.assert_called_once_with(
        ["pkg", "fetch", "-y", "-gd", "mutt"],
        output_loglevel="trace",
        python_shell=False,
    )


def test_fetch_with_regex_flag():
    """
    Test pkgng.fetch with enabled regex flag
    """

    targets = "mutt"
    pkg_cmd = MagicMock(return_value=targets)
    patches = {
        "cmd.run": pkg_cmd,
        "pkg_resource.stringify": MagicMock(),
        "pkg_resource.parse_targets": MagicMock(return_value=targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        pkgs = pkgng.fetch(targets, regex=True)
        assert pkgs == targets
    pkg_cmd.assert_called_once_with(
        ["pkg", "fetch", "-y", "-gx", "mutt"],
        output_loglevel="trace",
        python_shell=False,
    )


def test_fetch_with_fromrepo_flag():
    """
    Test pkgng.fetch with enabled fromrepo flag
    """

    targets = "mutt"
    pkg_cmd = MagicMock(return_value=targets)
    patches = {
        "cmd.run": pkg_cmd,
        "pkg_resource.stringify": MagicMock(),
        "pkg_resource.parse_targets": MagicMock(return_value=targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        pkgs = pkgng.fetch(targets, fromrepo="FreeBSD-poudriere")
        assert pkgs == targets
    pkg_cmd.assert_called_once_with(
        ["pkg", "fetch", "-y", "-r", "FreeBSD-poudriere", "-g", "mutt"],
        output_loglevel="trace",
        python_shell=False,
    )


def test_fetch_with_localcache_flag():
    """
    Test pkgng.fetch with enabled localcache flag
    """

    targets = "mutt"
    pkg_cmd = MagicMock(return_value=targets)
    patches = {
        "cmd.run": pkg_cmd,
        "pkg_resource.stringify": MagicMock(),
        "pkg_resource.parse_targets": MagicMock(return_value=targets),
    }
    with patch.dict(pkgng.__salt__, patches):
        pkgs = pkgng.fetch(targets, local=True)
        assert pkgs == targets
    pkg_cmd.assert_called_once_with(
        ["pkg", "fetch", "-y", "-gL", "mutt"],
        output_loglevel="trace",
        python_shell=False,
    )


def test_which_with_default_flags():
    """
    Test pkgng.which
    """
    which_cmd = MagicMock(
        return_value={
            "stdout": "/usr/local/bin/mutt was installed by package mutt-2.0.6",
            "retcode": 0,
        }
    )
    with patch.dict(pkgng.__salt__, {"cmd.run": which_cmd}):

        result = pkgng.which("/usr/local/bin/mutt")
        assert result
        which_cmd.assert_called_with(
            ["pkg", "which", "/usr/local/bin/mutt"],
            output_loglevel="trace",
            python_shell=False,
        )


def test_which_with_origin_flag():
    """
    Test pkgng.which with enabled origin flag
    """
    which_cmd = MagicMock(
        return_value={
            "stdout": "/usr/local/bin/mutt was installed by package mail/mutt",
            "retcode": 0,
        }
    )
    with patch.dict(pkgng.__salt__, {"cmd.run": which_cmd}):

        result = pkgng.which("/usr/local/bin/mutt", origin=True)
        assert result
        which_cmd.assert_called_with(
            ["pkg", "which", "-o", "/usr/local/bin/mutt"],
            output_loglevel="trace",
            python_shell=False,
        )



import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.dig as dig
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        dig: {
            "__salt__": {
                "cmd.run_all": cmdmod.run_all,
            },
        }
    }


def test_dig_cname_found():
    dig_mock = MagicMock(
        return_value={
            "pid": 2018,
            "retcode": 0,
            "stderr": "",
            "stdout": "bellanotte1986.github.io.",
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.CNAME("www.eitr.tech") == "bellanotte1986.github.io."


def test_dig_cname_none_found():
    dig_mock = MagicMock(
        return_value={
            "pid": 2022,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.CNAME("www.google.com") == ""



"""
    :codeauthor: :email:`Zach Moody <zmoody@do.co>`
"""
import pytest

import salt.modules.netbox as netbox
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase

try:
    import pynetbox  # pylint: disable=unused-import

    HAS_PYNETBOX = True
except ImportError:
    HAS_PYNETBOX = False


NETBOX_RESPONSE_STUB = {
    "device_name": "test1-router1",
    "url": "http://test/",
    "device_role": {"name": "router", "url": "http://test/"},
}


def mocked_clean_kwargs_filter(**kwargs):
    """
    Mocked args.clean_kwargs for filter tests
    """
    return {"site": "test"}


def mocked_clean_kwargs_get(**kwargs):
    """
    Mocked args.clean_kwargs for get tests
    """
    return {"name": "test"}


@pytest.mark.skipif(HAS_PYNETBOX is False, reason="pynetbox lib not installed")
@patch("salt.modules.netbox._config", MagicMock())
class NetBoxTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            netbox: {},
        }

    def test_get_by_id(self):
        with patch("pynetbox.api", MagicMock()) as mock:
            with patch.dict(
                netbox.__utils__, {"args.clean_kwargs": mocked_clean_kwargs_get}
            ):
                netbox.get_("dcim", "devices", id=1)
                self.assertEqual(mock.mock_calls[1], call().dcim.devices.get(1))

    def test_get_by_name(self):
        with patch("pynetbox.api", MagicMock()) as mock:
            with patch.dict(
                netbox.__utils__, {"args.clean_kwargs": mocked_clean_kwargs_get}
            ):
                netbox.get_("dcim", "devices", name="test")
                self.assertEqual(
                    mock.mock_calls[1], call().dcim.devices.get(name="test")
                )

    def test_filter_by_site(self):
        with patch("pynetbox.api", MagicMock()) as mock:
            with patch.dict(
                netbox.__utils__, {"args.clean_kwargs": mocked_clean_kwargs_filter}
            ):
                netbox.filter_("dcim", "devices", site="test")
                self.assertEqual(
                    mock.mock_calls[1], call().dcim.devices.filter(site="test")
                )

    def test_filter_url(self):
        strip_url = netbox._strip_url_field(NETBOX_RESPONSE_STUB)
        self.assertTrue(
            "url" not in strip_url and "url" not in strip_url["device_role"]
        )

    def test_get_secret(self):
        with patch("pynetbox.api", MagicMock()) as mock:
            with patch.dict(
                netbox.__utils__, {"args.clean_kwargs": mocked_clean_kwargs_get}
            ):
                netbox.get_("secrets", "secrets", name="test")
                self.assertTrue("token" and "private_key_file" in mock.call_args[1])

    def test_token_present(self):
        with patch("pynetbox.api", MagicMock()) as mock:
            with patch.dict(
                netbox.__utils__, {"args.clean_kwargs": mocked_clean_kwargs_get}
            ):
                netbox.get_("dcim", "devices", name="test")
                self.assertTrue("token" in mock.call_args[1])



"""
Management of Block Devices

A state module to manage blockdevices

.. code-block:: yaml


    /dev/sda:
      blockdev.tuned:
        - read-only: True

    master-data:
      blockdev.tuned:
        - name: /dev/vg/master-data
        - read-only: True
        - read-ahead: 1024


.. versionadded:: 2014.7.0
"""

import logging
import os
import os.path
import time

import salt.utils.path

__virtualname__ = "blockdev"

# Init logger
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load this module if the disk execution module is available
    """
    if "disk.tune" in __salt__:
        return __virtualname__
    return (
        False,
        "Cannot load the {} state module: disk execution module not found".format(
            __virtualname__
        ),
    )


def tuned(name, **kwargs):
    """
    Manage options of block device

    name
        The name of the block device

    opts:
      - read-ahead
          Read-ahead buffer size

      - filesystem-read-ahead
          Filesystem Read-ahead buffer size

      - read-only
          Set Read-Only

      - read-write
          Set Read-Write
    """

    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    kwarg_map = {
        "read-ahead": "getra",
        "filesystem-read-ahead": "getfra",
        "read-only": "getro",
        "read-write": "getro",
    }

    if not __salt__["file.is_blkdev"]:
        ret["comment"] = "Changes to {} cannot be applied. Not a block device. ".format(
            name
        )
    elif __opts__["test"]:
        ret["comment"] = "Changes to {} will be applied ".format(name)
        ret["result"] = None
        return ret
    else:
        current = __salt__["disk.dump"](name)
        changes = __salt__["disk.tune"](name, **kwargs)
        changeset = {}
        for key in kwargs:
            if key in kwarg_map:
                switch = kwarg_map[key]
                if current[switch] != changes[switch]:
                    if isinstance(kwargs[key], bool):
                        old = current[switch] == "1"
                        new = changes[switch] == "1"
                    else:
                        old = current[switch]
                        new = changes[switch]
                    if key == "read-write":
                        old = not old
                        new = not new
                    changeset[key] = "Changed from {} to {}".format(old, new)
        if changes:
            if changeset:
                ret["comment"] = "Block device {} successfully modified ".format(name)
                ret["changes"] = changeset
            else:
                ret["comment"] = "Block device {} already in correct state".format(name)
        else:
            ret["comment"] = "Failed to modify block device {}".format(name)
            ret["result"] = False
    return ret


def formatted(name, fs_type="ext4", force=False, **kwargs):
    """
    Manage filesystems of partitions.

    name
        The name of the block device

    fs_type
        The filesystem it should be formatted as

    force
        Force mke2fs to create a filesystem, even if the specified device is
        not a partition on a block special device. This option is only enabled
        for ext and xfs filesystems

        This option is dangerous, use it with caution.

        .. versionadded:: 2016.11.0
    """
    ret = {
        "changes": {},
        "comment": "{} already formatted with {}".format(name, fs_type),
        "name": name,
        "result": False,
    }

    if not os.path.exists(name):
        ret["comment"] = "{} does not exist".format(name)
        return ret

    current_fs = _checkblk(name)

    if current_fs == fs_type:
        ret["result"] = True
        return ret
    elif not salt.utils.path.which("mkfs.{}".format(fs_type)):
        ret["comment"] = "Invalid fs_type: {}".format(fs_type)
        ret["result"] = False
        return ret
    elif __opts__["test"]:
        ret["comment"] = "Changes to {} will be applied ".format(name)
        ret["result"] = None
        return ret

    __salt__["disk.format"](name, fs_type, force=force, **kwargs)

    # Repeat fstype check up to 10 times with 3s sleeping between each
    # to avoid detection failing although mkfs has succeeded
    # see https://github.com/saltstack/salt/issues/25775
    # This retry maybe superfluous - switching to blkid
    for i in range(10):

        log.info("Check blk fstype attempt %d of 10", i + 1)
        current_fs = _checkblk(name)

        if current_fs == fs_type:
            ret["comment"] = "{} has been formatted with {}".format(name, fs_type)
            ret["changes"] = {"new": fs_type, "old": current_fs}
            ret["result"] = True
            return ret

        if current_fs == "":
            log.info("Waiting 3s before next check")
            time.sleep(3)
        else:
            break

    ret["comment"] = "Failed to format {}".format(name)
    ret["result"] = False
    return ret


def _checkblk(name):
    """
    Check if the blk exists and return its fstype if ok
    """

    blk = __salt__["cmd.run"](
        ["blkid", "-o", "value", "-s", "TYPE", name], ignore_retcode=True
    )
    return "" if not blk else blk



import os

import pytest


@pytest.fixture(scope="module")
def install():
    pytest.helpers.clean_env()

    # Create a custom config
    pytest.helpers.custom_config()

    pytest.helpers.run_command([pytest.INST_BIN, "/S", "/custom-config=custom_conf"])
    yield
    pytest.helpers.clean_env()


def test_binaries_present(install):
    assert os.path.exists(rf"{pytest.INST_DIR}\ssm.exe")


def test_config_present(install):
    assert os.path.exists(rf"{pytest.DATA_DIR}\conf\minion")


def test_config_correct(install):
    # The config file should be the custom config, unchanged
    with open(rf"{pytest.REPO_DIR}\custom_conf") as f:
        expected = f.readlines()

    with open(rf"{pytest.DATA_DIR}\conf\minion") as f:
        result = f.readlines()

    assert result == expected



import os
import pathlib
import subprocess
import sys

import psutil
import pytest
import yaml
from pytestskipmarkers.utils import platform

pytestmark = [
    pytest.mark.skip_on_windows,
    pytest.mark.skip_on_darwin,
]


@pytest.fixture
def pkg_paths():
    """
    Paths created by package installs
    """
    paths = [
        "/etc/salt",
        "/var/cache/salt",
        "/var/log/salt",
        "/var/run/salt",
        "/opt/saltstack/salt",
    ]
    return paths


@pytest.fixture
def pkg_paths_salt_user():
    """
    Paths created by package installs and owned by salt user
    """
    return [
        "/etc/salt/cloud.deploy.d",
        "/var/log/salt/cloud",
        "/opt/saltstack/salt/lib/python{}.{}/site-packages/salt/cloud/deploy".format(
            *sys.version_info
        ),
        "/etc/salt/pki/master",
        "/etc/salt/master.d",
        "/var/log/salt/master",
        "/var/log/salt/api",
        "/var/log/salt/key",
        "/var/cache/salt/master",
        "/var/run/salt/master",
    ]


@pytest.fixture
def pkg_paths_salt_user_exclusions():
    """
    Exclusions from paths created by package installs and owned by salt user
    """
    paths = [
        "/var/cache/salt/master/.root_key"  # written by salt, salt-run and salt-key as root
    ]
    return paths


def test_salt_user_master(salt_master, install_salt):
    """
    Test the correct user is running the Salt Master
    """
    match = False
    for proc in psutil.Process(salt_master.pid).children():
        assert proc.username() == "salt"
        match = True

    assert match


def test_salt_user_home(install_salt):
    """
    Test the salt user's home is /opt/saltstack/salt
    """
    proc = subprocess.run(
        ["getent", "passwd", "salt"], check=False, capture_output=True
    )
    assert proc.returncode == 0
    home = ""
    try:
        home = proc.stdout.decode().split(":")[5]
    except:
        pass
    assert home == "/opt/saltstack/salt"


def test_salt_user_group(install_salt):
    """
    Test the salt user is in the salt group
    """
    proc = subprocess.run(["id", "salt"], check=False, capture_output=True)
    assert proc.returncode == 0
    in_group = False
    try:
        for group in proc.stdout.decode().split(" "):
            if "salt" in group:
                in_group = True
    except:
        pass
    assert in_group is True


def test_salt_user_shell(install_salt):
    """
    Test the salt user's login shell
    """
    proc = subprocess.run(
        ["getent", "passwd", "salt"], check=False, capture_output=True
    )
    assert proc.returncode == 0
    shell = ""
    shell_exists = False
    try:
        shell = proc.stdout.decode().split(":")[6].strip()
        shell_exists = pathlib.Path(shell).exists()
    except:
        pass
    assert shell_exists is True


def test_pkg_paths(
    install_salt, pkg_paths, pkg_paths_salt_user, pkg_paths_salt_user_exclusions
):
    """
    Test package paths ownership
    """
    salt_user_subdirs = []
    for _path in pkg_paths:
        pkg_path = pathlib.Path(_path)
        assert pkg_path.exists()
        for dirpath, sub_dirs, files in os.walk(pkg_path):
            path = pathlib.Path(dirpath)
            # Directories owned by salt:salt or their subdirs/files
            if (
                str(path) in pkg_paths_salt_user or str(path) in salt_user_subdirs
            ) and str(path) not in pkg_paths_salt_user_exclusions:
                assert path.owner() == "salt"
                assert path.group() == "salt"
                salt_user_subdirs.extend(
                    [str(path.joinpath(sub_dir)) for sub_dir in sub_dirs]
                )
                # Individual files owned by salt user
                for file in files:
                    file_path = path.joinpath(file)
                    if str(file_path) not in pkg_paths_salt_user_exclusions:
                        assert file_path.owner() == "salt"
            # Directories owned by root:root
            else:
                assert path.owner() == "root"
                assert path.group() == "root"
                for file in files:
                    file_path = path.joinpath(file)
                    # Individual files owned by salt user
                    if str(file_path) in pkg_paths_salt_user:
                        assert file_path.owner() == "salt"
                    else:
                        assert file_path.owner() == "root"
                        assert file_path.group() == "root"



"""
tests.pytests.integration.modules.test_event
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import time

import pytest
from saltfactories.utils import random_string

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


def test_fire_master(event_listener, salt_master, salt_minion, salt_call_cli):
    """
    Test firing an event on the master event bus
    """
    event_tag = random_string("salt/test/event/")
    start_time = time.time()
    ret = salt_call_cli.run(
        "event.fire_master", "event.fire_master: just test it!!!!", event_tag
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data is True

    event_pattern = (salt_master.id, event_tag)
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events
    for event in matched_events:
        assert event.data["id"] == salt_minion.id
        assert event.data["cmd"] == "_minion_event"
        assert "event.fire_master: just test it!!!!" in event.data["data"]


def test_event_fire(event_listener, salt_minion, salt_sub_minion, salt_cli):
    """
    Test firing an even on both test minions local event bus
    """
    event_tag = random_string("salt/test/event/")
    data = {"event.fire": "just test it!!!!"}
    for minion_tgt in (salt_minion.id, salt_sub_minion.id):
        start_time = time.time()
        ret = salt_cli.run(
            "event.fire", data=data, tag=event_tag, minion_tgt=minion_tgt
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data is True

        event_pattern = (minion_tgt, event_tag)
        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=30
        )
        assert matched_events.found_all_events
        for event in matched_events:
            assert event.data == data


def test_send(event_listener, salt_master, salt_minion, salt_call_cli):
    """
    Test sending an event to the master event bus
    """
    event_tag = random_string("salt/test/event/")
    data = {"event.fire": "just test it!!!!"}
    start_time = time.time()
    ret = salt_call_cli.run(
        "event.send",
        event_tag,
        data=data,
        with_grains=True,
        with_pillar=True,
        preload={"foo": "bar"},
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data is True

    event_pattern = (salt_master.id, event_tag)
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events
    for event in matched_events:
        assert event.data["id"] == salt_minion.id
        assert event.data["cmd"] == "_minion_event"
        assert "event.fire" in event.data["data"]
        assert event.data["foo"] == "bar"
        assert event.data["data"]["grains"]["test_grain"] == "cheese"
        assert event.data["data"]["pillar"]["ext_spam"] == "eggs"



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.grub_legacy
"""


import errno

import pytest

import salt.modules.grub_legacy as grub_legacy
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {grub_legacy: {}}


def test_version():
    """
    Test for Return server version from grub --version
    """
    mock = MagicMock(return_value="out")
    with patch.dict(grub_legacy.__salt__, {"cmd.run": mock}):
        assert grub_legacy.version() == "out"


def test_conf():
    """
    Test for Parse GRUB conf file
    """
    file_data = IOError(errno.EACCES, "Permission denied")
    with patch("salt.utils.files.fopen", mock_open(read_data=file_data)), patch.object(
        grub_legacy, "_detect_conf", return_value="A"
    ):
        pytest.raises(CommandExecutionError, grub_legacy.conf)

    file_data = salt.utils.stringutils.to_str("\n".join(["#", "A B C D,E,F G H"]))
    with patch("salt.utils.files.fopen", mock_open(read_data=file_data)), patch.object(
        grub_legacy, "_detect_conf", return_value="A"
    ):
        conf = grub_legacy.conf()
        assert conf == {"A": "B C D,E,F G H", "stanzas": []}, conf



"""
Validate the functions in the localfs cache
"""
import errno
import shutil

import pytest

import salt.cache.localfs as localfs
import salt.payload
import salt.utils.files
from salt.exceptions import SaltCacheError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {localfs: {}}


@pytest.fixture
def tmp_cache_file(tmp_path):
    with patch.dict(localfs.__opts__, {"cachedir": tmp_path}):
        localfs.store(bank="bank", key="key", data="payload data", cachedir=tmp_path)
        yield tmp_path


def _create_tmp_cache_file(self, tmp_dir):
    """
    Helper function that creates a temporary cache file using localfs.store. This
    is to used to create DRY unit tests for the localfs cache.
    """
    self.addCleanup(shutil.rmtree, tmp_dir)
    with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
        localfs.store(bank="bank", key="key", data="payload data", cachedir=tmp_dir)


# 'store' function tests: 5


def test_handled_exception_cache_dir():
    """
    Tests that a SaltCacheError is raised when the base directory doesn't exist and
    cannot be created.
    """
    with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
        with patch("tempfile.mkstemp", MagicMock(side_effect=Exception)):
            with pytest.raises(Exception):
                localfs.store(bank="", key="", data="", cachedir="")


def test_unhandled_exception_cache_dir():
    """
    Tests that a SaltCacheError is raised when the base directory doesn't exist and
    cannot be created.
    """
    with patch("os.makedirs", MagicMock(side_effect=OSError(1, ""))):
        with pytest.raises(SaltCacheError):
            localfs.store(bank="", key="", data="", cachedir="")


def test_store_close_mkstemp_file_handle():
    """
    Tests that the file descriptor that is opened by os.open during the mkstemp call
    in localfs.store is closed before calling salt.utils.files.fopen on the filename.

    This test mocks the call to mkstemp, but forces an OSError to be raised when the
    close() function is called on a file descriptor that doesn't exist.
    """
    with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
        with patch("tempfile.mkstemp", MagicMock(return_value=(12345, "foo"))):
            with pytest.raises(OSError):
                localfs.store(bank="", key="", data="", cachedir="")


def test_store_error_writing_cache():
    """
    Tests that a SaltCacheError is raised when there is a problem writing to the
    cache file.
    """
    with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
        with patch("tempfile.mkstemp", MagicMock(return_value=("one", "two"))):
            with patch("os.close", MagicMock(return_value=None)):
                with patch("salt.utils.files.fopen", MagicMock(side_effect=IOError)):
                    with pytest.raises(SaltCacheError):
                        localfs.store(bank="", key="", data="", cachedir="")


def test_store_success(tmp_cache_file):
    """
    Tests that the store function writes the data to the serializer for storage.
    """
    # Read in the contents of the key.p file and assert "payload data" was written
    with salt.utils.files.fopen(str(tmp_cache_file / "bank" / "key.p"), "rb") as fh_:
        for line in fh_:
            assert b"payload data" in line


# 'fetch' function tests: 3


def test_fetch_return_when_cache_file_does_not_exist():
    """
    Tests that the fetch function returns an empty dict when the cache key file
    doesn't exist.
    """
    with patch("os.path.isfile", MagicMock(return_value=False)):
        assert localfs.fetch(bank="", key="", cachedir="") == {}


def test_fetch_error_reading_cache():
    """
    Tests that a SaltCacheError is raised when there is a problem reading the cache
    file.
    """
    with patch("os.path.isfile", MagicMock(return_value=True)):
        with patch("salt.utils.files.fopen", MagicMock(side_effect=IOError)):
            with pytest.raises(SaltCacheError):
                localfs.fetch(bank="", key="", cachedir="")


def test_fetch_success(tmp_cache_file):
    """
    Tests that the fetch function is able to read the cache file and return its data.
    """
    assert "payload data" in localfs.fetch(
        bank="bank", key="key", cachedir=tmp_cache_file
    )


# # 'updated' function tests: 3


def test_updated_return_when_cache_file_does_not_exist():
    """
    Tests that the updated function returns None when the cache key file doesn't
    exist.
    """
    with patch("os.path.isfile", MagicMock(return_value=False)):
        assert localfs.updated(bank="", key="", cachedir="") is None


def test_updated_error_when_reading_mtime():
    """
    Tests that a SaltCacheError is raised when there is a problem reading the mtime
    of the cache file.
    """
    with patch("os.path.isfile", MagicMock(return_value=True)):
        with patch("os.path.getmtime", MagicMock(side_effect=IOError)):
            with pytest.raises(SaltCacheError):
                localfs.fetch(bank="", key="", cachedir="")


def test_updated_success(tmp_cache_file):
    """
    Test that the updated function returns the modification time of the cache file
    """
    with patch(
        "os.path.join", MagicMock(return_value=str(tmp_cache_file / "bank" / "key.p"))
    ):
        assert isinstance(
            localfs.updated(bank="bank", key="key", cachedir=tmp_cache_file), int
        )


# # 'flush' function tests: 4


def test_flush_key_is_none_and_no_target_dir():
    """
    Tests that the flush function returns False when no key is passed in and the
    target directory doesn't exist.
    """
    with patch("os.path.isdir", MagicMock(return_value=False)):
        assert localfs.flush(bank="", key=None, cachedir="") is False


def test_flush_key_provided_and_no_key_file_false():
    """
    Tests that the flush function returns False when a key file is provided but
    the target key file doesn't exist in the cache bank.
    """
    with patch("os.path.isfile", MagicMock(return_value=False)):
        assert localfs.flush(bank="", key="key", cachedir="") is False


def test_flush_success(tmp_cache_file):
    """
    Tests that the flush function returns True when a key file is provided and
    the target key exists in the cache bank.
    """
    with patch("os.path.isfile", MagicMock(return_value=True)):
        assert localfs.flush(bank="bank", key="key", cachedir=tmp_cache_file) is True


def test_flush_error_raised():
    """
    Tests that a SaltCacheError is raised when there is a problem removing the
    key file from the cache bank
    """
    with patch("os.path.isfile", MagicMock(return_value=True)):
        with patch("os.remove", MagicMock(side_effect=OSError)):
            with pytest.raises(SaltCacheError):
                localfs.flush(bank="", key="key", cachedir="/var/cache/salt")


# # 'list' function tests: 3


def test_list_no_base_dir():
    """
    Tests that the ls function returns an empty list if the bank directory
    doesn't exist.
    """
    with patch("os.path.isdir", MagicMock(return_value=False)):
        assert localfs.list_(bank="", cachedir="") == []


def test_list_error_raised_no_bank_directory_access():
    """
    Tests that a SaltCacheError is raised when there is a problem accessing the
    cache bank directory.
    """
    with patch("os.path.isdir", MagicMock(return_value=True)):
        with patch("os.listdir", MagicMock(side_effect=OSError)):
            with pytest.raises(SaltCacheError):
                localfs.list_(bank="", cachedir="")


def test_list_success(tmp_cache_file):
    """
    Tests the return of the ls function containing bank entries.
    """
    assert localfs.list_(bank="bank", cachedir=tmp_cache_file) == ["key"]


# # 'contains' function tests: 1


def test_contains(tmp_cache_file):
    """
    Test the return of the contains function when key=None and when a key
    is provided.
    """
    assert localfs.contains(bank="bank", key=None, cachedir=tmp_cache_file) is True
    assert localfs.contains(bank="bank", key="key", cachedir=tmp_cache_file) is True


def test_mix_of_utf8_and_non_utf8_can_be_round_tripped(tmp_cache_file):
    data = {
        # Any unicode, which ideally is invalid ascii.
        "unicode": "áéí",
        # Any bytes so long as they're not valid utf-8
        "bytes": b"\xfe\x99\x00\xff",
    }
    bank = "bank"
    key = "key"

    localfs.store(bank, key, data, tmp_cache_file)
    actual = localfs.fetch(bank, key, tmp_cache_file)

    assert data == actual



"""
SmartOS Metadata grain provider

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils, salt.module.cmdmod
:platform:      SmartOS

.. versionadded:: 2017.7.0

"""

import logging
import os

import salt.modules.cmdmod
import salt.utils.dictupdate
import salt.utils.json
import salt.utils.path
import salt.utils.platform

__virtualname__ = "mdata"
__salt__ = {
    "cmd.run": salt.modules.cmdmod.run,
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Figure out if we need to be loaded
    """
    ## collect mdata grains in a SmartOS zone
    if salt.utils.platform.is_smartos_zone():
        return __virtualname__
    ## collect mdata grains in a LX zone
    if salt.utils.platform.is_linux() and "BrandZ virtual linux" in os.uname():
        return __virtualname__
    return False


def _user_mdata(mdata_list=None, mdata_get=None):
    """
    User Metadata
    """
    grains = {}

    if not mdata_list:
        mdata_list = salt.utils.path.which("mdata-list")

    if not mdata_get:
        mdata_get = salt.utils.path.which("mdata-get")

    if not mdata_list or not mdata_get:
        return grains

    for mdata_grain in __salt__["cmd.run"](
        mdata_list, ignore_retcode=True
    ).splitlines():
        if mdata_grain.startswith("ERROR:"):
            log.warning("mdata-list returned an error, skipping mdata grains.")
            continue
        mdata_value = __salt__["cmd.run"](
            "{} {}".format(mdata_get, mdata_grain), ignore_retcode=True
        )

        if not mdata_grain.startswith("sdc:"):
            if "mdata" not in grains:
                grains["mdata"] = {}

            log.debug("found mdata entry %s with value %s", mdata_grain, mdata_value)
            mdata_grain = mdata_grain.replace("-", "_")
            mdata_grain = mdata_grain.replace(":", "_")
            grains["mdata"][mdata_grain] = mdata_value

    return grains


def _sdc_mdata(mdata_list=None, mdata_get=None):
    """
    SDC Metadata specified by there specs
    https://eng.joyent.com/mdata/datadict.html
    """
    grains = {}
    sdc_text_keys = [
        "uuid",
        "server_uuid",
        "datacenter_name",
        "hostname",
        "dns_domain",
        "alias",
    ]
    sdc_json_keys = [
        "resolvers",
        "nics",
        "routes",
    ]

    if not mdata_list:
        mdata_list = salt.utils.path.which("mdata-list")

    if not mdata_get:
        mdata_get = salt.utils.path.which("mdata-get")

    if not mdata_list or not mdata_get:
        return grains

    for mdata_grain in sdc_text_keys + sdc_json_keys:
        mdata_value = __salt__["cmd.run"](
            "{} sdc:{}".format(mdata_get, mdata_grain), ignore_retcode=True
        )
        if mdata_value.startswith("ERROR:"):
            log.warning(
                "unable to read sdc:%s via mdata-get, mdata grain may be incomplete.",
                mdata_grain,
            )
            continue

        if not mdata_value.startswith("No metadata for "):
            if "mdata" not in grains:
                grains["mdata"] = {}
            if "sdc" not in grains["mdata"]:
                grains["mdata"]["sdc"] = {}

            log.debug(
                "found mdata entry sdc:%s with value %s", mdata_grain, mdata_value
            )
            mdata_grain = mdata_grain.replace("-", "_")
            mdata_grain = mdata_grain.replace(":", "_")
            if mdata_grain in sdc_json_keys:
                grains["mdata"]["sdc"][mdata_grain] = salt.utils.json.loads(mdata_value)
            else:
                grains["mdata"]["sdc"][mdata_grain] = mdata_value

    return grains


def mdata():
    """
    Provide grains from the SmartOS metadata
    """
    grains = {}
    mdata_list = salt.utils.path.which("mdata-list")
    mdata_get = salt.utils.path.which("mdata-get")

    grains = salt.utils.dictupdate.update(
        grains, _user_mdata(mdata_list, mdata_get), merge_lists=True
    )
    grains = salt.utils.dictupdate.update(
        grains, _sdc_mdata(mdata_list, mdata_get), merge_lists=True
    )

    return grains



"""
Simple Smoke Tests for Connected Proxy Minion
"""
import logging

import pytest

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="Deltaproxy minions do not currently work on spawning platforms.",
    ),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def skip_on_tcp_transport(request):
    if request.config.getoption("--transport") == "tcp":
        pytest.skip("Deltaproxy under the TPC transport is not working. See #61367")


@pytest.fixture(params=pytest.helpers.proxy.delta_proxy_minion_ids())
def proxy_id(request, salt_delta_proxy, skip_on_tcp_transport):
    return request.param


@pytest.fixture
def proxy_ids(request, salt_delta_proxy, skip_on_tcp_transport):
    return pytest.helpers.proxy.delta_proxy_minion_ids()


def test_can_it_ping(salt_cli, proxy_id, proxy_ids):
    """
    Ensure the proxy can ping
    """
    ret = salt_cli.run("test.ping", minion_tgt=proxy_id)
    assert ret.data is True


def test_can_it_ping_all(salt_cli, proxy_ids):
    """
    Ensure the proxy can ping (all proxy minions)
    """
    ret = salt_cli.run("-L", "test.ping", minion_tgt=",".join(proxy_ids))
    for _id in proxy_ids:
        assert ret.data[_id] is True


def test_list_pkgs(salt_cli, proxy_id):
    """
    Package test 1, really just tests that the virtual function capability
    is working OK.
    """
    ret = salt_cli.run("pkg.list_pkgs", minion_tgt=proxy_id)
    assert "coreutils" in ret.data
    assert "apache" in ret.data
    assert "redbull" in ret.data


def test_list_pkgs_all(salt_cli, proxy_ids):
    """
    Ensure the proxy can ping (all proxy minions)
    """
    pkg_list = {"apache": "2.4", "coreutils": "1.0", "redbull": "999.99", "tinc": "1.4"}
    ret = salt_cli.run("-L", "pkg.list_pkgs", minion_tgt=",".join(proxy_ids))
    for _id in proxy_ids:
        assert ret.data[_id] == pkg_list


def test_install_pkgs(salt_cli, proxy_id):
    """
    Package test 2, really just tests that the virtual function capability
    is working OK.
    """

    ret = salt_cli.run("pkg.install", "thispkg", minion_tgt=proxy_id)
    assert ret.data["thispkg"] == "1.0"

    ret = salt_cli.run("pkg.list_pkgs", minion_tgt=proxy_id)

    assert ret.data["apache"] == "2.4"
    assert ret.data["redbull"] == "999.99"
    assert ret.data["thispkg"] == "1.0"


def test_install_pkgs_all(salt_cli, proxy_ids):
    """
    Ensure the proxy can ping (all proxy minions)
    """
    install_ret = salt_cli.run(
        "-L", "pkg.install", "thispkg", minion_tgt=",".join(proxy_ids)
    )
    list_ret = salt_cli.run("-L", "pkg.list_pkgs", minion_tgt=",".join(proxy_ids))

    for _id in proxy_ids:

        assert install_ret.data[_id]["thispkg"] == "1.0"

        assert list_ret.data[_id]["apache"] == "2.4"
        assert list_ret.data[_id]["redbull"] == "999.99"
        assert list_ret.data[_id]["thispkg"] == "1.0"


def test_remove_pkgs(salt_cli, proxy_id):
    ret = salt_cli.run("pkg.remove", "apache", minion_tgt=proxy_id)
    assert "apache" not in ret.data

    # reinstall
    ret = salt_cli.run("pkg.install", "apache", minion_tgt=proxy_id)


def test_remove_pkgs_all(salt_cli, proxy_ids):
    """
    Ensure the proxy can ping (all proxy minions)
    """
    ret = salt_cli.run("-L", "pkg.remove", "coreutils", minion_tgt=",".join(proxy_ids))

    for _id in proxy_ids:
        assert "coreutils" not in ret.data[_id]

    # reinstall
    salt_cli.run("-L", "pkg.install", "coreutils", minion_tgt=",".join(proxy_ids))


def test_upgrade(salt_cli, proxy_id):
    ret = salt_cli.run("pkg.upgrade", minion_tgt=proxy_id)
    assert ret.data["coreutils"]["new"] == "2.0"
    assert ret.data["redbull"]["new"] == "1000.99"


def test_service_list(salt_cli, proxy_id):
    ret = salt_cli.run("service.list", minion_tgt=proxy_id)
    assert "ntp" in ret.data


def test_service_stop(salt_cli, proxy_id):
    ret = salt_cli.run("service.stop", "ntp", minion_tgt=proxy_id)
    ret = salt_cli.run("service.status", "ntp", minion_tgt=proxy_id)
    assert ret.data is False


def test_service_start(salt_cli, proxy_id):
    ret = salt_cli.run("service.start", "samba", minion_tgt=proxy_id)
    ret = salt_cli.run("service.status", "samba", minion_tgt=proxy_id)
    assert ret.data is True


def test_service_get_all(salt_cli, proxy_id):
    ret = salt_cli.run("service.get_all", minion_tgt=proxy_id)
    assert ret.data
    assert "samba" in ret.data


def test_grains_items(salt_cli, proxy_id):
    ret = salt_cli.run("grains.items", minion_tgt=proxy_id)
    assert ret.data["kernel"] == "proxy"
    assert ret.data["kernelrelease"] == "proxy"


def test_state_apply(salt_master, salt_cli, tmp_path, proxy_id):
    test_file = tmp_path / "testfile"
    core_state = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        test_file
    )

    with salt_master.state_tree.base.temp_file("core.sls", core_state):
        ret = salt_cli.run("state.apply", "core", minion_tgt=proxy_id)
        for value in ret.data.values():
            assert value["result"] is True


@pytest.mark.slow_test
def test_state_highstate(salt_master, salt_cli, tmp_path, proxy_id):
    test_file = tmp_path / "testfile"
    top_sls = """
    base:
      '*':
        - core
        """

    core_state = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        test_file
    )

    with salt_master.state_tree.base.temp_file(
        "top.sls", top_sls
    ), salt_master.state_tree.base.temp_file("core.sls", core_state):
        ret = salt_cli.run("state.highstate", minion_tgt=proxy_id)
        for value in ret.data.values():
            assert value["result"] is True


def test_config_get(salt_cli, proxy_id):
    """
    Ensure the config module returns the right id
    when targeting deltaproxy managed proxy minions.
    """
    ret = salt_cli.run("config.get", "id", minion_tgt=proxy_id)
    assert ret.data == proxy_id


def test_schedule_list(salt_cli, proxy_id):
    """
    Ensure schedule.list works
    """
    ret = salt_cli.run("schedule.list", minion_tgt=proxy_id)
    assert ret.data == "schedule: {}\n"


def test_schedule_add_list(salt_cli, proxy_id):
    """
    Ensure schedule.add works
    """
    ret = salt_cli.run(
        "schedule.add", name="job1", function="test.ping", minion_tgt=proxy_id
    )
    assert "result" in ret.data
    assert ret.data["result"]

    assert "comment" in ret.data
    assert ret.data["comment"] == "Added job: job1 to schedule."

    assert "changes" in ret.data
    assert ret.data["changes"] == {"job1": "added"}

    _expected = """schedule:
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: true
"""
    ret = salt_cli.run("schedule.list", minion_tgt=proxy_id)
    assert ret.data == _expected

    # clean out the scheduler
    salt_cli.run("schedule.purge", minion_tgt=proxy_id)


def test_schedule_add_list_all(salt_cli, proxy_ids):
    """
    Ensure schedule.add works when targeting a single minion
    and that the others are not affected.
    """
    ret = salt_cli.run(
        "schedule.add", name="job2", function="test.ping", minion_tgt=proxy_ids[0]
    )
    assert "result" in ret.data
    assert ret.data["result"]

    ret = salt_cli.run("-L", "schedule.list", minion_tgt=",".join(proxy_ids))

    # check every proxy except the first one
    for _id in proxy_ids[1:]:
        assert ret.data[_id] == "schedule: {}\n"

    # clean out the scheduler
    salt_cli.run("-L", "schedule.purge", minion_tgt=",".join(proxy_ids))



import shutil

import pytest

import salt.loader
import salt.roster.ansible as ansible
from tests.support.mock import patch

pytestmark = [
    pytest.mark.skip_if_binaries_missing("ansible-inventory"),
]


@pytest.fixture
def roster_opts():
    return {"roster_defaults": {"passwd": "test123"}}


@pytest.fixture
def configure_loader_modules(temp_salt_master, roster_opts):
    opts = temp_salt_master.config.copy()
    utils = salt.loader.utils(opts, whitelist=["json", "stringutils", "ansible"])
    runner = salt.loader.runner(opts, utils=utils, whitelist=["salt"])
    return {
        ansible: {"__utils__": utils, "__opts__": roster_opts, "__runner__": runner}
    }


@pytest.fixture
def expected_targets_return():
    return {
        "host1": {
            "host": "host1",
            "passwd": "test123",
            "minion_opts": {
                "escape_pods": 2,
                "halon_system_timeout": 30,
                "self_destruct_countdown": 60,
                "some_server": "foo.southeast.example.com",
            },
        },
        "host2": {
            "host": "host2",
            "passwd": "test123",
            "minion_opts": {
                "escape_pods": 2,
                "halon_system_timeout": 30,
                "self_destruct_countdown": 60,
                "some_server": "foo.southeast.example.com",
            },
        },
        "host3": {
            "host": "host3",
            "passwd": "test123",
            "minion_opts": {
                "escape_pods": 2,
                "halon_system_timeout": 30,
                "self_destruct_countdown": 60,
                "some_server": "foo.southeast.example.com",
            },
        },
    }


@pytest.fixture
def expected_docs_targets_return():
    return {
        "home": {
            "passwd": "password",
            "sudo": "password",
            "host": "12.34.56.78",
            "port": 23,
            "user": "gtmanfred",
            "minion_opts": {"http_port": 80},
        },
        "salt.gtmanfred.com": {
            "passwd": "password",
            "sudo": "password",
            "host": "127.0.0.1",
            "port": 22,
            "user": "gtmanfred",
            "minion_opts": {"http_port": 80},
        },
    }


@pytest.fixture(scope="module")
def roster_dir(tmp_path_factory):
    dpath = tmp_path_factory.mktemp("roster")
    roster_py_contents = """
    #!/usr/bin/env python

    import json
    import sys

    inventory = {
        "usa": {"children": ["southeast"]},
        "southeast": {
            "children": ["atlanta", "raleigh"],
            "vars": {
                "some_server": "foo.southeast.example.com",
                "halon_system_timeout": 30,
                "self_destruct_countdown": 60,
                "escape_pods": 2,
            },
        },
        "raleigh": ["host2", "host3"],
        "atlanta": ["host1", "host2"],
    }
    hostvars = {"host1": {}, "host2": {}, "host3": {}}

    if "--host" in sys.argv:
        print(json.dumps(hostvars.get(sys.argv[-1], {})))
    if "--list" in sys.argv:
        print(json.dumps(inventory))
    """
    roster_ini_contents = """
    [atlanta]
    host1
    host2

    [raleigh]
    host2
    host3

    [southeast:children]
    atlanta
    raleigh

    [southeast:vars]
    some_server=foo.southeast.example.com
    halon_system_timeout=30
    self_destruct_countdown=60
    escape_pods=2

    [usa:children]
    southeast
    """
    roster_yaml_contents = """
    atlanta:
      hosts:
        host1:
        host2:
    raleigh:
      hosts:
        host2:
        host3:
    southeast:
      children:
        atlanta:
        raleigh:
      vars:
        some_server: foo.southeast.example.com
        halon_system_timeout: 30
        self_destruct_countdown: 60
        escape_pods: 2
    usa:
      children:
        southeast:
    """
    docs_ini_contents = """
    [servers]
    salt.gtmanfred.com ansible_ssh_user=gtmanfred ansible_ssh_host=127.0.0.1 ansible_ssh_port=22 ansible_ssh_pass='password' ansible_sudo_pass='password'

    [desktop]
    home ansible_ssh_user=gtmanfred ansible_ssh_host=12.34.56.78 ansible_ssh_port=23 ansible_ssh_pass='password' ansible_sudo_pass='password'

    [computers:children]
    desktop
    servers

    [computers:vars]
    http_port=80
    """
    docs_script_contents = """
    #!/bin/bash
    echo '{
        "servers": [
            "salt.gtmanfred.com"
        ],
        "desktop": [
            "home"
        ],
        "computers": {
            "hosts": [],
            "children": [
                "desktop",
                "servers"
            ],
            "vars": {
                "http_port": 80
            }
        },
        "_meta": {
            "hostvars": {
                "salt.gtmanfred.com": {
                    "ansible_ssh_user": "gtmanfred",
                    "ansible_ssh_host": "127.0.0.1",
                    "ansible_sudo_pass": "password",
                    "ansible_ssh_pass": "password",
                    "ansible_ssh_port": 22
                },
                "home": {
                    "ansible_ssh_user": "gtmanfred",
                    "ansible_ssh_host": "12.34.56.78",
                    "ansible_sudo_pass": "password",
                    "ansible_ssh_pass": "password",
                    "ansible_ssh_port": 23
                }
            }
        }
    }'
    """
    with pytest.helpers.temp_file(
        "roster.py", roster_py_contents, directory=dpath
    ) as py_roster:
        py_roster.chmod(0o755)
        with pytest.helpers.temp_file(
            "roster.ini", roster_ini_contents, directory=dpath
        ), pytest.helpers.temp_file(
            "roster.yml", roster_yaml_contents, directory=dpath
        ), pytest.helpers.temp_file(
            "roster-docs.ini", docs_ini_contents, directory=dpath
        ):
            with pytest.helpers.temp_file(
                "roster-docs.sh", docs_script_contents, directory=dpath
            ) as script_roster:
                script_roster.chmod(0o755)
                try:
                    yield dpath
                finally:
                    shutil.rmtree(str(dpath), ignore_errors=True)


@pytest.mark.parametrize(
    "which_value",
    [False, None],
)
def test_virtual_returns_False_if_ansible_inventory_doesnt_exist(which_value):
    with patch("salt.utils.path.which", autospec=True, return_value=which_value):
        assert ansible.__virtual__() == (False, "Install `ansible` to use inventory")


def test_ini(roster_opts, roster_dir, expected_targets_return):
    roster_opts["roster_file"] = str(roster_dir / "roster.ini")
    with patch.dict(ansible.__opts__, roster_opts):
        ret = ansible.targets("*")
        assert ret == expected_targets_return


def test_yml(roster_opts, roster_dir, expected_targets_return):
    roster_opts["roster_file"] = str(roster_dir / "roster.yml")
    with patch.dict(ansible.__opts__, roster_opts):
        ret = ansible.targets("*")
        assert ret == expected_targets_return


def test_script(roster_opts, roster_dir, expected_targets_return):
    roster_opts["roster_file"] = str(roster_dir / "roster.py")
    with patch.dict(ansible.__opts__, roster_opts):
        ret = ansible.targets("*")
        assert ret == expected_targets_return


def test_docs_ini(roster_opts, roster_dir, expected_docs_targets_return):
    roster_opts["roster_file"] = str(roster_dir / "roster-docs.ini")
    with patch.dict(ansible.__opts__, roster_opts):
        ret = ansible.targets("*")
        assert ret == expected_docs_targets_return


def test_docs_script(roster_opts, roster_dir, expected_docs_targets_return):
    roster_opts["roster_file"] = str(roster_dir / "roster-docs.sh")
    with patch.dict(ansible.__opts__, roster_opts):
        ret = ansible.targets("*")
        assert ret == expected_docs_targets_return



"""
Stores eauth tokens in the filesystem of the master. Location is configured by the master config option 'token_dir'
"""


import hashlib
import logging
import os

import salt.payload
import salt.utils.files
import salt.utils.path
import salt.utils.verify

log = logging.getLogger(__name__)

__virtualname__ = "localfs"


def mk_token(opts, tdata):
    """
    Mint a new token using the config option hash_type and store tdata with 'token' attribute set
    to the token.
    This module uses the hash of random 512 bytes as a token.

    :param opts: Salt master config options
    :param tdata: Token data to be stored with 'token' attribute of this dict set to the token.
    :returns: tdata with token if successful. Empty dict if failed.
    """
    hash_type = getattr(hashlib, opts.get("hash_type", "md5"))
    tok = str(hash_type(os.urandom(512)).hexdigest())
    t_path = os.path.join(opts["token_dir"], tok)
    temp_t_path = "{}.tmp".format(t_path)
    while os.path.isfile(t_path):
        tok = str(hash_type(os.urandom(512)).hexdigest())
        t_path = os.path.join(opts["token_dir"], tok)
    tdata["token"] = tok
    try:
        with salt.utils.files.set_umask(0o177):
            with salt.utils.files.fopen(temp_t_path, "w+b") as fp_:
                fp_.write(salt.payload.dumps(tdata))
        os.rename(temp_t_path, t_path)
    except OSError:
        log.warning('Authentication failure: can not write token file "%s".', t_path)
        return {}
    return tdata


def get_token(opts, tok):
    """
    Fetch the token data from the store.

    :param opts: Salt master config options
    :param tok: Token value to get
    :returns: Token data if successful. Empty dict if failed.
    """
    t_path = os.path.join(opts["token_dir"], tok)
    if not salt.utils.verify.clean_path(opts["token_dir"], t_path):
        return {}
    if not os.path.isfile(t_path):
        return {}
    try:
        with salt.utils.files.fopen(t_path, "rb") as fp_:
            tdata = salt.payload.loads(fp_.read())
            return tdata
    except OSError:
        log.warning('Authentication failure: can not read token file "%s".', t_path)
        return {}


def rm_token(opts, tok):
    """
    Remove token from the store.

    :param opts: Salt master config options
    :param tok: Token to remove
    :returns: Empty dict if successful. None if failed.
    """
    t_path = os.path.join(opts["token_dir"], tok)
    try:
        os.remove(t_path)
        return {}
    except OSError:
        log.warning("Could not remove token %s", tok)


def list_tokens(opts):
    """
    List all tokens in the store.

    :param opts: Salt master config options
    :returns: List of dicts (tokens)
    """
    ret = []
    for (dirpath, dirnames, filenames) in salt.utils.path.os_walk(opts["token_dir"]):
        for token in filenames:
            ret.append(token)
    return ret



"""
Error generator to enable integration testing of salt runner error handling

"""

import salt.utils.error


def error(name=None, message=""):
    """
    If name is None Then return empty dict

    Otherwise raise an exception with __name__ from name, message from message

    CLI Example:

    .. code-block:: bash

        salt-run error
        salt-run error.error name="Exception" message="This is an error."
    """
    ret = {}
    if name is not None:
        salt.utils.error.raise_error(name=name, message=message)
    return ret



import pytest
import requests

import salt.utils.http
from tests.support.mock import MagicMock, patch


def test_requests_session_verify_ssl_false(ssl_webserver, integration_files_dir):
    """
    test salt.utils.http.session when using verify_ssl
    """
    for verify in [True, False, None]:
        kwargs = {"verify_ssl": verify}
        if verify is None:
            kwargs.pop("verify_ssl")

        if verify is True or verify is None:
            with pytest.raises(requests.exceptions.SSLError) as excinfo:
                session = salt.utils.http.session(**kwargs)
                ret = session.get(ssl_webserver.url("this.txt"))
        else:
            session = salt.utils.http.session(**kwargs)
            ret = session.get(ssl_webserver.url("this.txt"))
            assert ret.status_code == 200


def test_session_ca_bundle_verify_false():
    """
    test salt.utils.http.session when using
    both ca_bunlde and verify_ssl false
    """
    ret = salt.utils.http.session(ca_bundle="/tmp/test_bundle", verify_ssl=False)
    assert ret is False


def test_session_headers():
    """
    test salt.utils.http.session when setting
    headers
    """
    ret = salt.utils.http.session(headers={"Content-Type": "application/json"})
    assert ret.headers["Content-Type"] == "application/json"


def test_session_ca_bundle():
    """
    test salt.utils.https.session when setting ca_bundle
    """
    fpath = "/tmp/test_bundle"
    patch_os = patch("os.path.exists", MagicMock(return_value=True))
    with patch_os:
        ret = salt.utils.http.session(ca_bundle=fpath)
    assert ret.verify == fpath



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.log_test
    ~~~~~~~~~~~~~~~~~~~

    Test salt's "hacked" logging
"""


import io
import logging

from salt._logging.handlers import StreamHandler
from salt._logging.impl import SaltLoggingClass
from tests.support.helpers import TstSuiteLoggingHandler


def test_issue_2853_regex_TypeError():
    # Now, python's logging logger class is ours.
    # Let's make sure we have at least one instance
    log = SaltLoggingClass(__name__)

    # Test for a format which includes digits in name formatting.
    log_format = "[%(name)-15s] %(message)s"
    handler = TstSuiteLoggingHandler(format=log_format)
    log.addHandler(handler)

    # Trigger TstSuiteLoggingHandler.__enter__
    with handler:
        # Let's create another log instance to trigger salt's logging class
        # calculations.
        try:
            SaltLoggingClass("{}.with_digits".format(__name__))
        except Exception as err:  # pylint: disable=broad-except
            raise AssertionError("No exception should have been raised: {}".format(err))

    # Remove the testing handler
    log.removeHandler(handler)

    # Test for a format which does not include digits in name formatting.
    log_format = "[%(name)s] %(message)s"
    handler = TstSuiteLoggingHandler(format=log_format)
    log.addHandler(handler)

    # Trigger TstSuiteLoggingHandler.__enter__
    with handler:
        # Let's create another log instance to trigger salt's logging class
        # calculations.
        try:
            SaltLoggingClass("{}.without_digits".format(__name__))
        except Exception as err:  # pylint: disable=broad-except
            raise AssertionError("No exception should have been raised: {}".format(err))

        # Remove the testing handler
        log.removeHandler(handler)


def test_exc_info_on_loglevel():
    def raise_exception_on_purpose():
        1 / 0  # pylint: disable=pointless-statement

    log = SaltLoggingClass(__name__)

    # Only stream2 should contain the traceback
    stream1 = io.StringIO()
    stream2 = io.StringIO()
    handler1 = StreamHandler(stream1)
    handler2 = StreamHandler(stream2)

    handler1.setLevel(logging.INFO)
    handler2.setLevel(logging.DEBUG)

    log.addHandler(handler1)
    log.addHandler(handler2)

    try:
        raise_exception_on_purpose()
    except ZeroDivisionError as exc:
        log.error(
            "Exception raised on purpose caught: ZeroDivisionError",
            exc_info_on_loglevel=logging.DEBUG,
        )

    try:
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream1.getvalue()
        )
        assert "Traceback (most recent call last)" not in stream1.getvalue()
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream2.getvalue()
        )
        assert "Traceback (most recent call last)" in stream2.getvalue()
    finally:
        log.removeHandler(handler1)
        log.removeHandler(handler2)

    # Both streams should contain the traceback
    stream1 = io.StringIO()
    stream2 = io.StringIO()
    handler1 = StreamHandler(stream1)
    handler2 = StreamHandler(stream2)

    handler1.setLevel(logging.INFO)
    handler2.setLevel(logging.DEBUG)

    log.addHandler(handler1)
    log.addHandler(handler2)

    try:
        raise_exception_on_purpose()
    except ZeroDivisionError as exc:
        log.error(
            "Exception raised on purpose caught: ZeroDivisionError",
            exc_info_on_loglevel=logging.INFO,
        )

    try:
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream1.getvalue()
        )
        assert "Traceback (most recent call last)" in stream1.getvalue()
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream2.getvalue()
        )
        assert "Traceback (most recent call last)" in stream2.getvalue()
    finally:
        log.removeHandler(handler1)
        log.removeHandler(handler2)

    # No streams should contain the traceback
    stream1 = io.StringIO()
    stream2 = io.StringIO()
    handler1 = StreamHandler(stream1)
    handler2 = StreamHandler(stream2)

    handler1.setLevel(logging.ERROR)
    handler2.setLevel(logging.INFO)

    log.addHandler(handler1)
    log.addHandler(handler2)

    try:
        raise_exception_on_purpose()
    except ZeroDivisionError as exc:
        log.error(
            "Exception raised on purpose caught: ZeroDivisionError",
            exc_info_on_loglevel=logging.DEBUG,
        )

    try:
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream1.getvalue()
        )
        assert "Traceback (most recent call last)" not in stream1.getvalue()
        assert (
            "Exception raised on purpose caught: ZeroDivisionError"
            in stream2.getvalue()
        )
        assert "Traceback (most recent call last)" not in stream2.getvalue()
    finally:
        log.removeHandler(handler1)
        log.removeHandler(handler2)
